#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class Michalewicz(AbstractSurface):

    def __init__(self, param_dim=2, m=10, noise=None):
        """Michalewicz function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            m (float): Parameter that defines the steepness of the valleys and ridges. A larger m leads to a more
            difficult search. Default is 10.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # "Certified Global Minima for a Benchmark of Difficult Optimization Problems":
        # https://pdfs.semanticscholar.org/6e82/d0cd7acbd55774c131af8941db6c2f84b7ec.pdf
        if self.param_dim == 2:
            params = [2.202906, 1.570796]
        elif self.param_dim == 3:
            params = [2.202906, 1.570796, 1.284992]
        elif self.param_dim == 4:
            params = [2.202906, 1.570796, 1.284992, 1.923058]
        elif self.param_dim == 5:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470]
        elif self.param_dim == 6:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470, 1.570796]
        elif self.param_dim == 7:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470, 1.570796, 1.454414]
        elif self.param_dim == 8:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470, 1.570796, 1.454414, 1.756087]
        elif self.param_dim == 9:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470, 1.570796, 1.454414, 1.756087, 1.655717]
        elif self.param_dim == 10:
            params = [2.202906, 1.570796, 1.284992, 1.923058, 1.720470, 1.570796, 1.454414, 1.756087, 1.655717, 1.5]
        else:
            return None

        # scale onto (0,1)
        params = np.array(params) / np.pi
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        # maxima at the corners
        return None

    def _run(self, params):
        params = np.array(params)
        params = np.pi * params  # rescale onto [0, pi]
        result = 0.
        for i, x in enumerate(params):
            result -= np.sin(x) * np.sin((i + 1) * (x ** 2) / np.pi) ** (2 * self.m)

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

from olympus.planners.wrapper_scipy import WrapperScipy
from olympus.objects.object_config import Config


class Slsqp(WrapperScipy):

    METHOD = 'slsqp'

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy/reference/optimize.minimize-slsqp.html#optimize-minimize-slsqp
    def __init__(self, goal='minimize',  disp=False, eps=1.4901161193847656e-8, ftol=1e-6, maxiter=15000,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """
        Sequential Least SQuares Programming (SLSQP) optimizers. SciPy implementation.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            disp (bool): Set to True to print convergence messages. If False, verbosity is ignored and set to 0.
            eps (float): Step size used for numerical approximation of the Jacobian.
            ftol (float): Precision goal for the value of f in the stopping criterion.
            maxiter (int): Maximum number of iterations.
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        """

        WrapperScipy.__init__(**locals())



#!/usr/bin/env python

# ===============================================================================

from olympus import Logger
from olympus.utils import generate_id

# ===============================================================================


class Database:
    """ generic database collector - can connect to multiple databases

    This class is intended to provide the interface to other modules within
    olympus.
    """

    def __init__(self, kind="sqlite", *args, **kwargs):
        self.dbs = {}
        self.active_db = None
        if not kind is None:
            self.add_db(kind, *args, **kwargs)

    def __iter__(self):
        for db in self.dbs.values():
            for campaign in db.get_campaigns():
                yield campaign

    def __repr__(self):
        return self.db.__repr__()

    @property
    def file_types(self):
        file_types = []
        for db in self.dbs.values():
            file_types.extend(db.file_types)
        return file_types

    @property
    def is_processing(self):
        return self.active_db.is_processing

    @property
    def db(self):
        return self.active_db

    def is_valid_file_type(self, file_type):
        return self.db._validate(file_type=file_type)

    def _guess_db_kind(self, file_name):
        from . import databases

        file_type = file_name.split(".")[-1]
        for db_kind, db in databases.items():
            if db.is_valid_file_type(file_type):
                break
        else:
            from . import db_types

            Logger.log(
                "Could not find database type {}. Please choose from {}".format(
                    file_type, db_types
                ),
                "ERROR",
            )
            return None
        return db_kind

    def from_file(self, file_name, kind=None):
        """ connects to a database stored on disk

        Args:
            file_name (str): path and name of the database file

        Returns:
            (Database): loaded database
        """
        if kind is None:
            kind = self._guess_db_kind(file_name)
        if kind is None:
            Logger.log(
                "Please provide database format (automatic guessing failed)", "ERROR"
            )
            return None
        db_name = file_name.split("/")[-1].split(".")[0]
        if db_name in self.dbs:
            self.set_active_db(db_name)
        else:
            self.add_db(kind=kind, name=db_name)
            self.set_active_db(db_name)
        self.db.from_file(file_name)
        return self

    def set_active_db(self, db_name):
        self.active_db = self.dbs[db_name]

    def add_db(self, kind, *args, **kwargs):
        try:
            database = __import__(
                f"olympus.databases.database_{kind}", fromlist=[f"Wrapper_{kind}"]
            )
        except ModuleNotFoundError:
            Logger.log(" ... proceeding with pickle database", "INFO", only_once=True)
            kind = "pickle"
            database = __import__(
                f"olympus.databases.database_{kind}", fromlist=[f"Wrapper_{kind}"]
            )

        database = getattr(database, f"Wrapper_{kind}")
        db = database(*args, **kwargs)
        self.dbs[db.name] = db
        if self.active_db is None:
            self.active_db = db

    def get_campaigns(self):
        campaigns = []
        for db in self.dbs.values():
            campaigns.extend(db.get_campaigns())
        return campaigns

    def list_campaigns(self):
        campaign_ids = []
        for db in self.dbs.values():
            campaign_ids.extend(db.list_campaigns())
        return campaign_ids

    def log_campaign(self, campaign):
        self.active_db.log_campaign(campaign)


# ===============================================================================



#!/usr/bin/env python

import time
import numpy as np

from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner
from olympus.planners.utils_planner import get_init_guess, get_bounds
from olympus.utils    import daemon

#===============================================================================

class SteepestDescent(AbstractPlanner):

    def __init__(self, goal='minimize', learning_rate=1e-3, dx=1e-5, random_seed=None,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """

        Args:
            goal:
            learning_rate:
            dx:
            random_seed:
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        """

        AbstractPlanner.__init__(**locals())
        self.has_optimizer = False

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.bounds      = get_bounds(param_space)
        if self.init_guess is None:
            self.init_guess = get_init_guess(param_space, method=self.init_guess_method, random_seed=self.init_guess_seed)

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, params):
        params = self._project_into_domain(params)
        self.SUBMITTED_PARAMS.append(params)
        while len(self.RECEIVED_VALUES) == 0:
            time.sleep(0.1)
        value = self.RECEIVED_VALUES.pop(0)
        return value

    @daemon
    def start_optimizer(self):
        guess   = self.init_guess.copy()
        while True:
            func    = self._priv_evaluator(guess)
            dy      = np.zeros(len(guess))
            perturb = guess.copy()
            for index in range(len(guess)):
                perturb[index] += self.dx
                probed = self._priv_evaluator(perturb)
                dy[index] = (probed - func) / self.dx
                perturb[index] -= self.dx
            guess = guess - self.eta * dy
            guess = self._project_into_domain(domain)

    def _ask(self):
        if self.has_optimizer is False:
            self.start_optimizer()
            self.has_optimizer = True

        while len(self.SUBMITTED_PARAMS) == 0:
            print('SUBMITTED_PARAMS', len(self.SUBMITTED_PARAMS))
            time.sleep(0.1)
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector().from_array(params, self.param_space)



#===============================================================================

if __name__ == '__main__':

    from olympus import Parameter, ParameterSpace
    param_space = ParameterSpace()
    param_space.add(Parameter(name='param_0'))
    param_space.add(Parameter(name='param_1'))

    planner = SteepestDescent(learning_rate=1e-3, dx=1e-5, random_seed=None, init_guess=None)
    planner.set_param_space(param_space=param_space)
    param = planner.ask()
    print('PARAM', param)



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class AckleyPath(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Ackley path function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        # minimum at the centre
        params = [0.5] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        return None

    def _run(self, params):
        params = np.array(params)
        params = 64 * np.array(params) - 32  # rescale onto [-32, 32]
        a = 20.
        b = 0.2
        c = 2 * np.pi
        n = float(len(params))
        params = np.array(params)
        result = - a * np.exp(- b * np.sqrt(np.sum(params ** 2) / n)) - np.exp(np.sum(np.cos(c * params)) / n) + a + np.exp(1.)

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

import numpy as np
from olympus import Logger
from abc import abstractmethod
from olympus.objects import Object, abstract_attribute, ABCMeta, ParameterVector
from olympus.campaigns.param_space import ParameterSpace, ParameterContinuous


class AbstractSurface(Object, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self._create_param_space(self.param_dim)
        self._create_value_space()

    @abstract_attribute
    def param_dim(self):
        pass

    @abstract_attribute
    def minima(self):
        pass

    @abstract_attribute
    def maxima(self):
        pass

    @abstractmethod
    def _run(self, features):
        pass

    @property
    def kind(self):
        return type(self).__name__

    def _create_param_space(self, param_dim):
        """
        Creates as continuous k-dimensional unit hypercube parameter space, where
        k = `param_dim`. In this space, the surface will live.
        """
        self.param_space = ParameterSpace()
        for i in range(param_dim):
            param = ParameterContinuous(name=f"param_{i}", low=0.0, high=1.0)
            self.param_space.add(param)

    def _create_value_space(self, value_dim=1):
        self.value_space = ParameterSpace()
        for _ in range(value_dim):
            value = ParameterContinuous(name=f"value_{_}", low=0.0, high=1.0)
            self.value_space.add(value)

    def run(self, params, return_paramvector=False):
        """Evaluate the surface at the chosen location.

        Args:
            params (array): Set of input parameters for which to return the function value.
            return_paramvector (bool): Whether to return a ``ParameterVector`` object instead of a list of lists.
                Default is False.

        Returns:
            values (ParameterVector): function values evaluated at the chosen locations.
        """
        if isinstance(params, float) or isinstance(params, int):
            params = np.array([params])
        elif type(params) == list:
            params = np.array(params)
        if len(params.shape) == 1:
            params = np.expand_dims(params, axis=0)

        # TODO: these validations could be moved to ParameterSpace class
        # validate params
        if params.shape[1] != len(self.param_space):
            message = (f'Dimensions of provided params ({params.shape[1]}) does not match expected '
                       f'dimension ({len(self.param_space)})')
            Logger.log(message, 'ERROR')

        # this raises warnings for out-of-bounds parameters
        for param_set in params:
            self.param_space.validate(param_set)

        # get values from the surface class
        y_preds = [[self._run(param_set)] for param_set in params]  # 2d array

        # if we are not asking for a ParamVector, we can just return y_preds
        if return_paramvector is False:
            return y_preds

        # return a ParameterVector
        # NOTE: while we do not allow batches or multiple objectives yet, this code is supposed to be able to support
        #  those
        y_pred_objects = []  # list of ParamVectors with all samples and objectives
        # iterate over all samples (if we returned a batch of predictions)
        for y_pred in y_preds:
            y_pred_object = ParameterVector()
            # iterate over all objectives/targets
            for target_name, y in zip(['target_0'], y_pred):
                y_pred_object.from_dict({target_name: y})
            # append object to list
            y_pred_objects.append(y_pred_object)

        return y_pred_objects



#!/usr/bin/env python

# ===============================================================================

import traceback

# ===============================================================================

from olympus import Logger
from .abstract_plotter import AbstractPlotter


# ===============================================================================
# check for matplotlib

Plotter = None
check_further = False

# ===============================================================================
# check for matplotlib

try:
    import matplotlib
except ModuleNotFoundError:
    error = traceback.format_exc()
    for line in error.split("\n"):
        if "ModuleNotFoundError" in line:
            module = line.strip().strip("'").split("'")[-1]
    message = """Plotter requires {module}, which could not be found.
    Install {module} for prettier plots""".format(
        module=module
    )
    Logger.log(message, "WARNING", only_once=True)
else:
    from .plotter_matplotlib import PlotterMatplotlib

    Plotter = PlotterMatplotlib
    check_further = True

    from .olympus_colors import get_olympus_colors

# ===============================================================================
# check for seaborn

if check_further:
    try:
        import seaborn
    except ModuleNotFoundError:
        error = traceback.format_exc()
        for line in error.split("\n"):
            if "ModuleNotFoundError" in line:
                module = line.strip().strip("'").split("'")[-1]
        message = """Plotter requires {module}, which could not be found.
        Please install {module} to use the plotter""".format(
            module=module
        )
        Logger.log(message, "WARNING", only_once=True)
    else:
        from .plotter_seaborn import PlotterSeaborn

        Plotter = PlotterSeaborn



#!/usr/bin/env python

import numpy as np
from olympus.objects.abstract_object import Object
from olympus.datasets.dataset import Dataset
from olympus import Logger
from functools import reduce


class DataTransformer(Object):

    def __init__(self, transformations='identity'):
        """Applies transformations to the columns of a 2d array.

        Args:
            transformations (str, list): Desired transformation. If a list is provided, the transformations will
                be performed one after the other one as provided in the list. Note that the `periodic` transformation
                is currently supported only as the first transformation, and only as a forward transform.
                Transformation options are:
                    `identity`: identity transform.
                    `normalize`: features are scaled to be between 0 and 1.
                    `standardize`: features are centered at 0 with standard deviation of 1.
                    `mean`: features are divided by their mean.
                    `sqrt_mean`: the square root of the features is taken after having been divided by their mean.
                    `periodic`: takes periodic variables in a Dataset and splits them into a sine and cosine representation.
        """
        Object.__init__(self)

        # check the input arguments
        if isinstance(transformations, str):
            transformations = [transformations]
        self._validate_args(transformations)

        # select transformations
        # note that we select the back transformations in reversed order
        self.transformations = transformations
        self._transforms      = [getattr(self, f'_forward_{trans}') for trans in self.transformations]
        self._back_transforms = [getattr(self, f'_backward_{trans}') for trans in reversed(self.transformations)]

        # the stats we will need
        self._mean = None
        self._stddev = None
        self._min = None
        self._max = None
        self._periodic_info = {}  # each key is a column index, each value is low/high

        # other info
        self.trained = False
        self._dims = None

    # ====================
    # Methods for the user
    # ====================
    def train(self, data):
        """Computes the statistics (e.g. mean and standard deviation) needed for the chosen transformation from the
        provided dataset. With the exception of the 'identity' transform, the DataTransformer always needs to be
        trained before the `transform` and `back_transform` methods can be used.

        Args:
            data (array, Dataset): the data used to compute the statistics needed for the transformation. This can
                be a 2d numpy array, or a Dataset object.
        """

        self._dims = None  # reset _dims if we retrain the DataTransformer

        # for splitting periodic variables we need a dataset, so that we can check which variables are periodic
        # and that are their lower/upper bounds
        if 'periodic' in self.transformations:
            if isinstance(data, Dataset) is False:
                message = 'in order to transform periodic variables you need to provide a Dataset object as the data argument'
                Logger.log(message, 'ERROR')

            # remember the input dimensions
            self._dims = np.shape(data.data)[1]

            # extract the info about periodic variables
            self._parse_dataset_for_periodic(data)

            # Now swap dataset for data after periodic transform. This is done just in case the periodic transform is
            # composed with other transformations that will then require operating on a higher dimensional array
            # the means, stddev etc. statistics will need to have matching dimensions
            data = self._forward_periodic(data.data.to_numpy())
        else:
            # allow passing a dataset
            if isinstance(data, Dataset) is True:
                data = data.data.to_numpy()
            self._validate_data(data)
            # remember the input dimensions
            self._dims = np.shape(data)[1]

        # ------------------------
        # Get stats about the data
        # ------------------------
        data = np.array(data)

        self._mean = np.mean(data, axis=0)
        self._stddev = np.std(data, axis=0)
        self._min = np.amin(data, axis=0)
        self._max = np.amax(data, axis=0)

        self.trained = True

    def transform(self, data):
        """Performs the transform on the provided data.

        Args:
            data (ndarray): a 2-dimensional numpy array where each row is a sample and each column a feature.
                `numpy.shape(data)` should return (num_samples, num_features)'

        Returns
            data_transformed (ndarray): the transformed data.

        """
        if isinstance(data, Dataset) is True:
            data = data.data.to_numpy()
        else:
            data = np.array(data)
        self._validate_data(data)

        if self.trained is False and self.transformation != 'identity':
            raise ValueError(f'DataTransformer needs to be trained before the transformation {self.transformation} '
                             f'can be applied.')

        pipeline = self._compose_transformations(self._transforms)
        return pipeline(data)

    def back_transform(self, data):
        """Performs the inverse transform on the provided data.

        Args:
           data (ndarray): a 2-dimensional numpy array where each row is a sample and each column a feature.
                `numpy.shape(data)` should return (num_samples, num_features)'

        Returns
            data_transformed (ndarray): the transformed data.

        """
        if isinstance(data, Dataset) is True:
            data = data.data.to_numpy()
        else:
            data = np.array(data)
        self._validate_data(data)

        if self.trained is False and self.transformation != 'identity':
            raise ValueError(f'DataTransformer needs to be trained before the transformation {self.transformation} '
                             f'can be applied.')

        pipeline = self._compose_transformations(self._back_transforms)
        return pipeline(data)

    # ===============
    # Private Methods
    # ===============
    def _forward_standardize(self, data):
        return (data - self._mean) / self._stddev

    def _backward_standardize(self, data):
        return data * self._stddev + self._mean

    def _forward_normalize(self, data):
        return (data - self._min) / (self._max - self._min)

    def _backward_normalize(self, data):
        return (self._max - self._min) * data + self._min

    def _forward_identity(self, data):
        return data

    def _backward_identity(self, data):
        return data

    def _forward_mean(self, data):
        return data / self._mean

    def _backward_mean(self, data):
        return data * self._mean

    def _forward_log_mean(self, data):
        return np.log(data / self._mean + 1e-4)

    def _backward_log_mean(self, data):
        return (np.exp(data) - 1e-4) * self._mean

    def _forward_sqrt_mean(self, data):
        return np.sqrt(data / self._mean)

    def _backward_sqrt_mean(self, data):
        return np.square(data) * self._mean

    def _forward_periodic(self, data):
        projections = []
        # iterate over columns in data
        for i, col in enumerate(data.T):
            if i in self._periodic_info:
                low = self._periodic_info[i][0]
                high = self._periodic_info[i][1]
                cosine = np.cos(2 * np.pi * (col - low) / (high - low))
                sine = np.sin(2 * np.pi * (col - low) / (high - low))
                projections.append(cosine)
                projections.append(sine)
            else:
                projections.append(col)

        return np.array(projections).T

        # TODO: if we want to allow a backward transform, we will need to store the position of sin/cos for all
        #  variables/columns transformed

    def _backward_periodic(self, data):
        raise NotImplementedError

    def _parse_dataset_for_periodic(self, dataset):
        # TODO/FIXME: this assumes the target is always the last column in a Dataset!!
        for i, param in enumerate(dataset.param_space):
            if param.is_periodic is True:
                # each key is a column index, each value is low/high
                self._periodic_info[i] = [param.low, param.high]

    def _validate_data(self, data):
        # check a 2-dimensional object is provided
        if data.ndim != 2:
            raise ValueError('Incorrect data format provided. Please provide a 2-dimensional array where each row is '
                             'a sample and each column a feature. `numpy.shape(data)` should return (num_samples, '
                             'num_features)')

        # check we have the same number of features as before
        if self._dims is not None:
            if np.shape(data)[1] != self._dims:
                if 'periodic' in self.transformations:
                    raise AssertionError('periodic back transformations are not yet supported')
                else:
                    raise AssertionError('dimensionality mismatch')

    @staticmethod
    def _validate_args(transformations):
        # check validity of transformation argument
        for transformation in transformations:
            if not (hasattr(DataTransformer, f'_forward_{transformation}') and hasattr(DataTransformer, f'_backward_{transformation}')):
                raise NotImplementedError(f'transformation {transformation} not implemented. Please select one of the '
                                          f'available transformation.')

        if 'periodic' in transformations and transformations.index('periodic') != 0:
            message = 'periodic transform is allowed only as the first transformation'
            Logger.log(message, 'ERROR')

    @staticmethod
    def _compose_transformations(functions_list):
        # compose transformations as provided in the list
        return lambda x: reduce(lambda a, f: f(a), functions_list, x)



#!/usr/bin/env python

from olympus import Emulator, Surface, Logger
from olympus.campaigns import Campaign
from olympus.objects   import Object

#===============================================================================

class Evaluator(Object):

    def __init__(self, planner, emulator=None, surface=None, campaign=Campaign(), database=None):
        """ The Evaluator does higher level operations that Planners and 
        Emulators do not do on their own. For instance, communicating parameters 
        and measurements to each other, keeping track of them ensuring they 
        match, and storing these in a Campaign object. All this can also be done 
        by the user using planner, emulator and campaign objects, which might 
        allow more customization. However, Evaluator provides a convenient 
        higher-level interface for common optimization tasks.

        Args:
            planner (Planner): an instance of a Planner.
            emulator (Emulator): an instance of a trained Emulator.
            surface (Surface): an instance of a Surface
            campaign (Campaign): an instance of a Campaign. By default, a new 
                Campaign instance is created. If this is set to None, no campaign 
                info will be stored.
            database (object): ...
        """
        Object.__init__(**locals())

        if emulator is not None:
            assert surface is None
            self.emulator_type = 'numerical'
        elif surface is not None:
            assert emulator is None 
            self.emulator_type = 'analytic'
            self.emulator = surface
        else:
            Logger.log('One of emulator or surface needs to be provided', 'FATAL')

#        if isinstance(self.emulator, Emulator):
#            self.emulator_type = 'numerical'
#        elif isinstance(self.emulator, Surface):
#            self.emulator_type = 'analytic'

        # provide the planner with the parameter space.
        # NOTE: right now, outside of Evaluator, the param_space for a planner 
        #       needs to be set "manually" by the user
        self.planner.set_param_space(self.emulator.param_space)  # param space in emulator as it originates from dataset

        if self.campaign is not None:
            self.campaign.set_planner_specs(planner)
            self.campaign.set_emulator_specs(emulator)


    def optimize(self, num_iter=1):
        """Optimizes a surface for a fixed number of iterations.

        Args:
            num_iter (int): Maximum number of iterations allowed.
        """

        # Optimize: i.e. call the planner recommend method for max_iter times
        for i in range(num_iter):
            # get new params from planner
            # NOTE: now we get 1 param at a time, a possible future expansion is 
            #       to return batches
            params = self.planner.recommend(observations=self.campaign.observations)

            # get measurement from emulator/surface
            values = self.emulator.run(params.to_array(), return_paramvector=True)

            # store parameter and measurement pair in campaign
            if self.campaign is not None:
                self.campaign.add_observation(params, values)

            # if we have a database, log the campaign status
            if self.database is not None:
                self.database.log_campaign(self.campaign)



#!/usr/bin/env python

from olympus.utils.misc import check_planner_module

# ----------------
# Import planner
# ----------------

planner = 'Scipy'
module  = 'scipy'
link    = 'https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html'
check_planner_module(planner, module, link)

# ------------------------
# Import Planner Wrapper
# ------------------------

from .wrapper_scipy import WrapperScipy



#!/usr/bin/env python

from abc import abstractmethod
from olympus import Object


class AbstractModel(Object):

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)

    @abstractmethod
    def train(self, X, y):
        """Method to train the model. We expect some standard output from this method.
        """
        pass

    @abstractmethod
    def predict(self, X):
        """Method that returns a prediction. We expect some standard output from this method.
        """
        pass



#!/usr/bin/env python

#===============================================================================

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_context('paper', font_scale = 1.5)

from olympus.plotter import AbstractPlotter

#===============================================================================

class PlotterSeaborn(AbstractPlotter):

    def _set_color_palette(self, line_theme = 'deep'):
        # NOTE/WARNING: the number of individual elements in this color palette
        # should be increased when more planners are added ...
        self.line_palette = sns.color_palette(line_theme, 20)

    def _plot(self, emulators, planners, measurements, file_name = None, show = False):

        self._set_color_palette()
        num_plots  = len(emulators)
        num_graphs = len(planners)
        fig = plt.figure(figsize = (6, 5 * num_plots))
        axs = []
        for plot_index in range(num_plots):
            ax = plt.subplot2grid((num_plots, 1), (plot_index, 0))
            axs.append(ax)

        for plot_index, emulator in enumerate(emulators):
            ax = axs[plot_index]
            for planner_ix, planner in enumerate(planners):
                measurements[emulator][planner]['vals'] = np.squeeze(measurements[emulator][planner]['vals'])
                sns.lineplot(
                    x = 'idxs',
                    y = 'vals',
                    data = measurements[emulator][planner],
                    color = 'k',
                    ax = ax,
                    linewidth = 5,
                    ci = None
                )
                sns.lineplot(
                    x    = 'idxs',
                    y    = 'vals',
                    data = measurements[emulator][planner],
                    ax   = ax,
                    linewidth = 4,
                    color = self.line_palette[planner_ix],
                    label = planner
                )
            ax.grid(linestyle = ':')
            ax.set_title(f'{emulator.capitalize()}')

            ax.set_xlabel('# evaluations')
            ax.set_ylabel('measurement')

        #plt.legend(loc='upper right', fontsize=12)
        plt.legend(fontsize = 12)
        plt.tight_layout()

        if file_name is not None:
            plt.savefig(file_name, bbox_inches = 'tight')
        if show is True:
            plt.show()



#!/usr/bin/env python

#======================================================================

from olympus import Logger

from olympus.objects import Object

#======================================================================

class ObjectParameter(Object):

    ATT_NAME = {'type': 'string', 'default': 'parameter'}

#======================================================================

class ObjectParameterContinuous(ObjectParameter):

    ATT_TYPE        = {'type': 'string', 'default': 'continuous', 'valid': ['continuous']}
    ATT_LOW         = {'type': 'float',  'default': 0.0}
    ATT_HIGH        = {'type': 'float',  'default': 1.0}
    ATT_IS_PERIODIC = {'type': 'bool',   'default': False}

    def __str__(self):
        return f"Continuous (name='{self.name}', low={self.low}, high={self.high}, is_periodic={self.is_periodic})"

    def _validate(self):
        return self.low < self.high

#======================================================================

class Parameter(ObjectParameter):

    KINDS = {'continuous': ObjectParameterContinuous}

    def __init__(self, kind = 'continuous', **kwargs):
        if kind in self.KINDS:
            self.kind = kind
            for prop in dir(self.KINDS[kind]):
                if prop.startswith('ATT_'):
                    setattr(self, prop, getattr(self.KINDS[kind], prop))
            self.KINDS[kind].__init__(self)
            for key, value in kwargs.items():
                if 'ATT_{}'.format(key.upper()) in dir(self):
                    self.add(key, value)
            if not self.KINDS[kind]._validate(self):
                message = 'Could not validate {}'.format(str(self))
                Logger.log(message, 'WARNING')
        else:
            message = '''Could not initialize parameter.
            Parameter kind {} is unknown. Please choose from {}'''.format(kind, ','.join(list(self.KINDS.keys())))
            Logger.log(message, 'ERROR')


    def __str__(self):
        return self.KINDS[self.kind].__str__(self)

#======================================================================



#!/usr/bin/env python

import numpy as np

from olympus import Object
from olympus.objects import ParameterVector


# ==========
# Main Class
# ==========
class Observations:
    def __init__(self):
        self.params = None  # expected to be a list of numpy arrays
        self.values = None  # expected to be a list of numpy arrays

        self.param_space = None
        self.value_space = None

        self._params_as_vectors = None
        self._values_as_vectors = None

    def _params_append(self, param):
        if self.params is None:
            self.params = np.array([param])
        else:
            self.params = np.append(self.params, [param], axis=0)

    def _values_append(self, value):
        if self.values is None:
            self.values = np.array([value])
        else:
            self.values = np.append(self.values, [value], axis=0)

    def set_param_space(self, param_space):
        self.param_space = param_space

    def set_value_space(self, value_space):
        self.value_space = value_space

    def _guess_param_space(self, param):
        self.param_space = param.param_space

    def _guess_value_space(self, value):
        self.value_space = value.param_space

    def add_observation(self, param, value):
        if isinstance(param, list):
            for param_ in param:
                if isinstance(param_, ParameterVector):
                    self._params_append(param_.to_array())
                    if self.param_space is None:
                        self._guess_param_space(param_)
                else:
                    self._params_append(param)
        else:
            if isinstance(param, ParameterVector):
                self._params_append(param.to_array())
                if self.param_space is None:
                    self._guess_param_space(param)
            else:
                self._params_append(param)

        if isinstance(value, list):
            for value_ in value:
                if isinstance(value_, ParameterVector):
                    self._values_append(value_.to_array())
                    if self.value_space is None:
                        self._guess_value_space(value_)
                else:
                    self._values_append(value)
        else:
            if isinstance(value, ParameterVector):
                self._values_append(value.to_array())
                if self.value_space is None:
                    self._guess_value_space(value)
            else:
                self._values_append(value)

        # reset caches
        self._params_as_vectors = None
        self._values_as_vectors = None

    def _construct_param_vectors(self):
        if self.params is None:
            self._params_as_vectors = []
        else:
            self._params_as_vectors = np.array(
                [
                    ParameterVector(param, param_space=self.param_space)
                    for param in self.params
                ]
            )

    def _construct_value_vectors(self):
        if self.values is None:
            self._values_as_vectors = []
        else:
            self._values_as_vectors = np.array(
                [
                    ParameterVector(value, param_space=self.value_space)
                    for value in self.values
                ]
            )

    def get_params(self, as_array=True):
        if self.params is None:
            return []

        if as_array is True:
            return np.array(self.params)
        elif as_array is False:
            if self._params_as_vectors is None:
                self._construct_param_vectors()
            return np.array([param.to_dict() for param in self._params_as_vectors])
        else:
            NotImplementedError

    def get_values(self, as_array=True, opposite=False):
        if self.values is None:
            return []

        if as_array is True:
            if opposite is False:
                return np.array(self.values)
            elif opposite is True:
                return -np.array(self.values)
        elif as_array is False:
            if self._values_as_vectors is None:
                self._construct_value_vectors()
            if opposite is False:
                return [value.to_dict() for value in self._values_as_vectors]
            elif opposite is True:
                values_dict = [value.to_dict() for value in self._values_as_vectors]
                for value_dict in values_dict:
                    for key, value in value_dict.items():
                        value_dict[key] = -1 * value
                return np.array(values_dict)

    def __OLD__init__(self):
        """ Class that contains all parameters and measurements collected from planners and emulators, i.e. `observed`
        parameters and measurements.
        """
        self.params = []  # expected to be a list of ParamVector objects
        self.values = []  # expected to be a list of ParamVector objects

    def __OLD_add_observation(self, param, value):
        """

        Args:
            param:
            value:

        Returns:

        """
        if isinstance(param, list):
            self.params.extend(param)
        elif isinstance(param, ParameterVector):
            self.params.append(param)

        if isinstance(value, list):
            self.values.extend(value)
        elif isinstance(param, ParameterVector):
            self.values.append(value)
        else:
            raise NotImplementedError

    def __OLD_get_params(self, as_array=True):
        """

        Args:
            as_array:

        Returns:

        """
        if as_array is True:
            return np.array([param.to_array() for param in self.params])
        elif as_array is False:
            return np.array([param.to_dict() for param in self.params])

    def __OLD_get_values(self, as_array=True, opposite=False):
        """

        Args:
            as_array (bool): Whether to return the values as an array.
            opposite (bool): Whether to return the opposite of the values. Default is False.

        Returns:

        """

        # NOTE: we use flipped values for the measurements to allow maximization in the planners!

        # return as array
        if as_array is True:
            # return normal values
            if opposite is False:
                return np.array([value.to_array() for value in self.values])
            # return flipped values
            elif opposite is True:
                return np.array([-1 * value.to_array() for value in self.values])
        # return as dict
        elif as_array is False:
            # return normal values
            if opposite is False:
                return np.array([value.to_dict() for value in self.values])
            # return flipped values
            elif opposite is True:
                values_dict = [value.to_dict() for value in self.values]
                for value_dict in values_dict:
                    for key, value in value_dict.items():
                        value_dict[key] = -1 * value
                return np.array(values_dict)



#!/usr/bin/env python

import time
import numpy as np 

from olympus          import __scratch__
from olympus.planners import AbstractPlanner
from olympus.objects  import ParameterVector
from olympus.utils    import daemon

#===============================================================================

class Smac(AbstractPlanner):


	def __init__(self, goal='minimize', rng=None):
		AbstractPlanner.__init__(**locals())
		self.seed = 100691
		self.rng = np.random.RandomState()
		self.has_optimizer = False 
		self.is_converged  = False
		self.budget = 10**8
		self.SUBMITTED_PARAMS = []
		self.RECEIVED_VALUES  = []
		self._params = np.array([[]])
		self._values = np.array([[]])

	def _set_param_space(self, param_space):
		self.param_space = param_space

		from ConfigSpace.hyperparameters import UniformFloatHyperparameter 
		from smac.configspace            import ConfigurationSpace 
		from smac.optimizer.objective    import average_cost
		from smac.runhistory.runhistory  import RunHistory
		from smac.scenario.scenario      import Scenario
		from smac.stats.stats            import Stats	
		from smac.utils.io.traj_logging  import TrajLogger	

		self.cs = ConfigurationSpace()
		for param in param_space:
			if param.type == 'continuous':
				var = UniformFloatHyperparameter(param.name, param.low, param.high)
				self.cs.add_hyperparameter(var)
		self.runhistory = RunHistory(aggregate_func=average_cost)
		self.scenario = Scenario({
			'run_obj': 'quality', 'runcount-limit': self.budget, 'cs': self.cs})
		self.stats = Stats(self.scenario)
		self.traj_logger = TrajLogger(output_dir=__scratch__, stats=self.stats)


	def _set_observations(self, observations):
		from smac.tae.execute_ta_run import StatusType
		self._params = observations.get_params(as_array=True)
		self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
		if len(self._values) > 0:
			self.RECEIVED_VALUES.append(self._values[-1])
			self.runhistory.add(self.smac_param, self._values[-1], time=1., status=StatusType.SUCCESS)


	def _priv_evaluator(self, params):
		return None


	def create_optimizer(self):
		from smac.epm.rf_with_instances import RandomForestWithInstances
		from smac.initial_design.default_configuration_design import DefaultConfiguration
		from smac.intensification.intensification import Intensifier
		from smac.optimizer.smbo import SMBO 
		from smac.optimizer.acquisition import EI
		from smac.optimizer.ei_optimization import InterleavedLocalAndRandomSearch
		from smac.optimizer.objective import average_cost
		from smac.runhistory.runhistory2epm import RunHistory2EPM4Cost
		from smac.tae.execute_ta_run import StatusType
		from smac.utils.constants import MAXINT
		from smac.utils.util_funcs import get_types


		TAE_RUNNER = self._priv_evaluator


		runhistory2epm = RunHistory2EPM4Cost(scenario=self.scenario, num_params=len(self.param_space),
                                           	 success_states=[StatusType.SUCCESS, StatusType.CRASHED],
                                             impute_censored_data=False, impute_state=None)

		intensifier = Intensifier(tae_runner=TAE_RUNNER, stats=self.stats, traj_logger=self.traj_logger,
                                  rng=self.rng, instances=self.scenario.train_insts, cutoff=self.scenario.cutoff,
                                  deterministic=self.scenario.deterministic, run_obj_time=self.scenario.run_obj == "runtime",
                                  always_race_against=self.scenario.cs.get_default_configuration() \
                                        if self.scenario.always_race_default else None,
                                  instance_specifics=self.scenario.instance_specific,
                                  minR=self.scenario.minR, maxR=self.scenario.maxR)

		types, bounds = get_types(self.scenario.cs, self.scenario.feature_array)
		model = RandomForestWithInstances(types=types, bounds=bounds, seed=self.rng.randint(MAXINT), 
			instance_features=self.scenario.feature_array, pca_components=self.scenario.PCA_DIM)
		acq_func = EI(model=model)

		smbo_args = {
			'scenario': self.scenario,
			'stats': self.stats,
			'initial_design': DefaultConfiguration(tae_runner=TAE_RUNNER, scenario=self.scenario, stats=self.stats, traj_logger=self.traj_logger, rng=self.rng),
			'runhistory': self.runhistory,
			'runhistory2epm': runhistory2epm,
			'intensifier': intensifier,
			'aggregate_func': average_cost,
			'num_run': self.seed,
			'model': model,
			'acq_optimizer': InterleavedLocalAndRandomSearch(acq_func, self.scenario.cs, np.random.RandomState(seed=self.rng.randint(MAXINT))),
			'acquisition_func': acq_func,
			'rng': self.rng,
			'restore_incumbent': None,
		}

		self.smbo = SMBO(**smbo_args)


	def _generate(self):
		if self.has_optimizer is False:
			self.create_optimizer()
			self.has_optimizer = True 

		smac_param_list = self.smbo.choose_next(self._params, self._values)
		for smac_param in smac_param_list:
			self.smac_param = smac_param 
			break

		param_dict = {key: self.smac_param[key] for key in self.smac_param}
		return ParameterVector().from_dict(param_dict)

#===============================================================================



#!/usr/bin/env python

#===============================================================================

import os
import json

from olympus import Logger
from abc import ABCMeta as NativeABCMeta


# ====================
# Olympus Object class
# ====================
class Object:

    """ Abstraction of a dictionary

    - facilitates declaration of defaults
    - can be extended to enable further modifications
    """

    def __init__(self, me = '', indent = 0, **kwargs):
        ''' creates empty object and loads defaults

        Args:
            me     (str): arbitrary name to identify the object
            indent (int): number of spaces used in string
                representation
        '''
        self.me           = me
        self.indent       = indent

        self.props        = []
        self.attrs        = []
        self.max_prop_len = 0

        # get defaults
        for prop in dir(self):
            if prop.startswith('ATT_'):
                attr = getattr(self, prop).copy()
                default = attr['default']
                if callable(default):
                    default = attr['default']()
                self.add(prop[4:].lower(), default)

        # overwrite defaults with any kwargs we have received
        for key, value in kwargs.items():
            if 'ATT_{}'.format(key.upper()) in dir(self) or not key in self:
                self.add(key, value)
            else:
                message = f'Attribute {key} is already defined for {self}'
                Logger.log(message, 'WARNING')

        valid = self._validate()
        if not valid:
            message = 'Could not validate {}'.format(str(self))
            Logger.log(message, 'WARNING')

    def __iter__(self):
        for _, prop in enumerate(self.props):
            yield prop, self.attrs[_]

    def __contains__(self, prop):
        return prop in self.props


    def __getattr__(self, prop):
        prop_name = prop[4:]
        if prop.startswith('set_'):
            def set_attr(attr):
                setattr(self, prop_name, attr)
            return set_attr
        elif prop.startswith('get_'):
            self.add(prop, lambda: self.get(prop_name))
            return getattr(self, prop)
        elif prop == '__getstate__':
            return lambda: self.__dict__
        elif prop == '__setstate__':
            def set_state(_dict):
                self.__dict__ = _dict
            return set_state
        elif f'ATT_{prop.upper()}' in dir(self):
            return getattr(self, f'ATT_{prop.upper()}')['default']
        raise AttributeError(f"Object has no attribute {prop}")


    def __getitem__(self, prop):
        return getattr(self, prop)


    def __setitem__(self, prop, attr):
        self.add(prop, attr)


    def __str__(self):
        string = '{}\n'.format(self.me)
        indent = ' '*self.indent
        for prop in sorted(self.props):
            string += '{ind}--> {prop: <{width}}{attr}\n'.format(
                    ind   = indent,
                    prop  = prop + ':',
                    width = self.max_prop_len + 2,
                    attr  = getattr(self, prop))
        return string[:-1]


    @property
    def defaults(self):
        defaults = []
        for prop in dir(self):
            if prop.startswith('ATT_'):
                attr_dict = getattr(self, prop)
                attr_dict['name'] = prop[4:]
                defaults.append(attr_dict)
        return defaults


    def _validate(self, **kwargs):
        for key, value in kwargs.items():
            if not value in getattr(self, 'ATT_{}'.format(key.upper()))['valid']:
                return False
        return True


    def add(self, prop, attr):
        ''' dynamically adds property and attribute to object

        Args:
            prop (any): property associated with attribute
            attr (any): property value
        '''

        if prop in ['props', 'attrs']:
            return

        setattr(self, prop, attr)
        if prop in self.props:
            self.attrs[self.props.index(prop)] = attr
        else:
            self.props.append(prop)
            self.attrs.append(attr)
            self.max_prop_len = max(len(prop), self.max_prop_len)


    def get(self, prop):
        ''' returns attribute associated with given property

        Args:
            prop (any): valid property

        Returns:
            any: attribute associated with property
        '''
        return getattr(self, prop)


    def update(self, prop, attr):
        self.add(prop, attr)


    def from_dict(self, info_dict):
        ''' returns object representation of given dictionary

        Args:
            info_dict (dict): dictionary to be represented

        Returns:
            Object: Object representation of dictionary
        '''
        for key, value in info_dict.items():
            self.add(key, value)
        return self


    def to_dict(self, exclude = []):
        ''' returns dictionary representation of presented object

        Args:
            exclude (list of any): properties to be excluded

        Returns:
            dict: representation of presented object
        '''
        info_dict = {}
        for prop in sorted(self.props):
            if prop in exclude: continue
            # ATTENTION: perhaps check for callables instead
            if prop.startswith('get_'): continue
            if prop.startswith('set_'): continue

            attr = getattr(self, prop)
            if isinstance(attr, Object):
                attr = attr.to_dict()
            if isinstance(attr, list):
                converted = []
                for elem in attr:
                    if isinstance(elem, Object):
                        elem = elem.to_dict()
                    converted.append(elem)
                attr = converted
            info_dict[prop] = attr
        return info_dict


    def from_json(self, json_file='config.json'):
        if os.path.isfile(json_file):
            with open(json_file, 'r') as content:
                config = json.loads(content.read())
        else:
            config = json.loads(json_file)
        self.from_dict(info_dict=config)
        return self


    def to_json(self, json_file='config.json'):
        config_dict = self.to_dict()
#        with open(json_file, 'w') as fp:
        config_json = json.dumps(config_dict, sort_keys=True, indent=4)
        return config_json


    def reset(self):
        self.props = []
        self.attrs = []


# ============================
# Abstract Attribute Decorator
# ============================
def abstract_attribute(obj=None):
    if obj is None:
        obj = DummyAttribute()
    obj.__is_abstract_attribute__ = True
    return obj


class DummyAttribute:
    pass


class ABCMeta(NativeABCMeta):

    def __call__(cls, *args, **kwargs):
        instance = NativeABCMeta.__call__(cls, *args, **kwargs)
        abstract_attributes = {name for name in dir(instance)
                               if getattr(getattr(instance, name), '__is_abstract_attribute__', False)}
        if abstract_attributes:
            raise NotImplementedError("Can't instantiate {0} without abstract "
                                      "attributes: {1}".format(cls.__name__, ', '.join(abstract_attributes)))
        return instance



#!/usr/bin/env python 

#======================================================================

from olympus.objects.abstract_object  import Object, abstract_attribute, ABCMeta

from olympus.objects.object_parameter import ObjectParameter
from olympus.objects.object_parameter import ObjectParameterContinuous as ParameterContinuous 
from olympus.objects.object_parameter import Parameter

from olympus.objects.object_objective import ObjectObjective           as Objective

from olympus.objects.object_parameter_vector import ObjectParameterVector as ParameterVector
from olympus.objects.object_config import Config
#======================================================================



#!/usr/bin/env python

import numpy as np

from olympus import Object

# ==============================================================================


class Analyzer(Object):
    def __init__(self, campaigns=[]):
        Object.__init__(**locals())

    def _get_best_vals(self, campaigns):
        vals_ = [campaign.best_values for campaign in campaigns]
        max_len = np.amax([len(val_) for val_ in vals_])
        vals = np.zeros((len(vals_), max_len)) + np.nan
        for _, val_ in enumerate(vals_):
            vals[_, : len(val_)] = np.squeeze(val_[:, 0])
        return vals

    def set_campaigns(self, campaigns):
        self.campaigns = campaigns

    def _get_best_reduction(self, operator, campaigns=None, locs=None, ci_method=None, ci_size=100):
        ''' computes statistics on the provided campaigns

        Statistics are computed following the specific reduction operation. Uncertainty
        estimates can be requested by choosing a method to determine confidence
        intervals and specifying the desired interval range (0-100)

        Args:
            operator (): reduction operator
            campaigns (list of campaign): campaigns for which to compute statistics
            locs (array of floats): locations at which statistics should be computed
            ci_method (str): method to estimate uncertainties (choose from: None, "bootstrap")
            ci_size (int): size of confidence interval (0-100)

        Returns:
            array: computed statistics
        '''
        if campaigns is None:
            campaigns = self.campaigns
        best_vals = self._get_best_vals(campaigns)
        if locs is None:
            best_vals_red = operator(best_vals, axis=0)
        else:
            locs = np.array(locs)
            locs = locs[np.where(locs < best_vals.shape[1])[0]]
            best_vals_red = operator(best_vals, axis=0)[locs - 1]

        if ci_method is None:
            return best_vals_red

        if ci_method == 'bootstrap':

            # implementing bootstrapping
            NUM_BOOTS = 100

            # samples = (# num_boot, # campaigns, length of trace)
            idxs    = np.random.randint(low=0, high=best_vals.shape[0], size=(NUM_BOOTS, best_vals.shape[0], best_vals.shape[1]))
            samples = []
            for _ in range(NUM_BOOTS):
                sample = [best_vals[idxs[_, :, __], __] for __ in range(best_vals.shape[1])]
                samples.append(sample)
            samples = np.array(samples)
            reductions = operator(samples, axis=2)

            lower_reductions = np.empty(best_vals.shape[1])
            upper_reductions = np.empty(best_vals.shape[1])

            for _ in range(best_vals.shape[1]):
                sorted_reductions = np.sort(reductions[:, _])
                lower_reductions[_] = sorted_reductions[int( (100 - ci_size) / 2 * (NUM_BOOTS-1) / 100)]
                upper_reductions[_] = sorted_reductions[int( (ci_size - (100 - ci_size) / 2) * (NUM_BOOTS-1) /  100)]

            return best_vals_red, {'lower': lower_reductions, 'upper': upper_reductions}

        else:
            raise NotImplementedError


    def get_best_mean(self, campaigns=None, locs=None, ci_method=None, ci_size=100):
        return self._get_best_reduction(np.nanmean, campaigns, locs, ci_method=ci_method, ci_size=ci_size)

    def get_best_median(self, campaigns=None, locs=None):
        return self._get_best_reduction(np.nanmedian, campaigns, locs)

    def get_best_std(self, campaigns=None, locs=None):
        return self._get_best_reduction(np.nanstd, campaigns, locs)

    def get_best_min(self, campaigns=None, locs=None):
        return self._get_best_reduction(np.nanmin, campaigns, locs)

    def get_best_max(self, campaigns=None, locs=None):
        return self._get_best_reduction(np.nanmax, campaigns, locs)



#!/usr/bin/env python


from olympus import Logger

from .abstract_object import Object


class Config(Object):

    def __init__(self, from_dict=None, from_json=None, name='CustomConfig'):
        """

        Args:
            from_dict:
            from_json:
            name:
        """
        super(Config, self).__init__(me=name)
        self.name = name

        if from_dict is not None:
            self.from_dict(info_dict=from_dict)
        elif from_json is not None:
            self.from_json(json_file=from_json)

        if from_dict is not None and from_json is not None:
            message = 'you have passed both "from_dict" and "from_json" arguments to Config: "from_json" will be discarded'
            Logger.log(message, 'WARNING')

    def __repr__(self):
        return f"<Config (name={self.me})>"

    def __str__(self):
        # determine spacing - additional numbers is to ensure a minimum spacing due to the headers
        max_prop_len = max([self.max_prop_len, 9])
        max_attr_len = max([len(str(att)) for att in self.attrs] + [5])
        max_type_len = max([len(str(type(att).__name__)) for att in self.attrs] + [4])
        total = max_prop_len + max_attr_len + max_type_len + 6

        # Title
        string = '=' * total + '\n'
        ind = int(round((total - len(self.me) - 8) / 2, 0))
        string += ' ' * ind + f'Config: {self.me}\n'
        string += '=' * total + '\n'

        # Headers
        string += '{0:<{3}}{1:<{4}}{2:<{5}}\n'.format('Parameter', 'Value', 'Type',
                                                      max_prop_len + 2,
                                                      max_attr_len + 2,
                                                      max_type_len + 2)
        string += '-' * total + '\n'

        # Parameters/Arguments
        for prop in sorted(self.props):
            string += '{0:<{3}}{1:<{4}}{2:<{5}}\n'.format(prop,
                                                          str(getattr(self, prop)),
                                                          str(type(getattr(self, prop)).__name__),
                                                          max_prop_len + 2,
                                                          max_attr_len + 2,
                                                          max_type_len + 2)
        string += '=' * total + '\n'
        return string



#!/usr/bin/env python

from olympus import Logger

from . import get_noises_list
from . import import_noise
from . import AbstractNoise


def Noise(kind='GaussianNoise', scale=1):
    """Convenience function to access noise objects. It returns a certain noise object
    with defaults arguments.

    Args:
        kind (str or AbstractNoise): Keyword identifying one of the noise objects available in Olympus. Alternatively,
            you can pass a custom noise class that is a subclass of AbstractNoise.

    Returns:
        Noise (object): An instance of the chosen noise.
    """
    _validate_noise_kind(kind)
    # if a string is passed, then load the corresponding wrapper
    if type(kind) == str:
        noise = import_noise(kind)
        noise = noise(scale=scale)
    # if an instance of a planner is passed, simply return the same instance
    elif isinstance(kind, AbstractNoise):
        noise = kind
    # if a custom class is passed, then that is the 'wrapper'
    elif issubclass(kind, AbstractNoise):
        noise = kind()

    return noise


def _validate_noise_kind(kind):
    # if we received a string
    if type(kind) == str:
        avail_noises = get_noises_list()
        if kind not in avail_noises:
            message = ('Noise "{0}" not available in Olympus. Please choose '
                       'from one of the available noise objects: {1}'.format(kind, ', '.join(avail_noises)))
            Logger.log(message, 'FATAL')

    # if we get an instance of a noise class
    elif isinstance(kind, AbstractNoise):
        # make sure it has the necessary methods
        for method in ['_add_noise']:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, 'FATAL')

    # if we received a custom planner class
    elif issubclass(kind, AbstractNoise):
        # make sure it has the necessary methods
        for method in ['_add_noise']:
            implementation = getattr(kind, method, None)
            if not callable(implementation):
                message = f'The object {kind} does not implement the necessary method "{method}"'
                Logger.log(message, 'FATAL')

    # if we do not know what was passed raise an error
    else:
        message = 'Could not initialize Noise: the argument "kind" is neither a string or AbstractNoise subclass'
        Logger.log(message, 'FATAL')



#!/usr/bin/env python

import os, sys, glob
__home__ = os.path.dirname(os.path.abspath(__file__))


class ModelLoader:

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._find_models()

    def __getattr__(self, attr):
        if attr in ['Model', 'AbstractModel'] or attr.startswith('Wrapper'):
            attr_file = ModelLoader.class_to_file(attr)
            module = __import__(f'olympus.models.{attr_file}', fromlist=[attr])
            _class = getattr(module, attr)
            setattr(self, attr, _class)
            return _class
        else:
            model = ModelLoader.import_model(attr)
            setattr(self, attr, model)
            return model

    @staticmethod
    def class_to_file(class_name):
        file_name = class_name[0].lower()
        for character in class_name[1:]:
            if character.isupper():
                file_name += f'_{character.lower()}'
            else:
                file_name += character
        return file_name

    @staticmethod
    def file_to_class(file_name):
        class_name = file_name[0].upper()
        next_upper = False
        for character in file_name[1:]:
            if character == '_':
                next_upper = True
                continue
            if next_upper:
                class_name += character.upper()
            else:
                class_name += character
            next_upper = False
        return class_name

    @staticmethod
    def import_model(attr):
        attr_file = ModelLoader.class_to_file(attr)
        module = __import__(f'olympus.models.model_{attr_file}', fromlist=[attr])
        _class = getattr(module, attr)
        return _class

    def _find_models(self):
        self.model_files = []
        self.model_names = []
        self.model_map  = {}
        for dir_name in glob.glob(f'{__home__}/model_*'):
            model_name = dir_name.split('/')[-1][6:]
            self.model_files.append(model_name)
            self.model_names.append(ModelLoader.file_to_class(model_name))

    def get_models_list(self):
        return sorted(self.model_names)


sys.modules[__name__] = ModelLoader(**locals())



#!/usr/bin/env python

#======================================================================

from olympus.campaigns.param_space  import ParameterSpace
from olympus.campaigns.observations import Observations
from olympus.campaigns.campaign     import Campaign


#!/usr/bin/env python

from olympus.utils.misc import check_planner_module

#===============================================================================

planner = 'ParticleSwarms'
module  = 'pyswarms'
link    = 'https://pyswarms.readthedocs.io/en/latest/'
check_planner_module(planner, module, link)

#===============================================================================

from .wrapper_particle_swarms import ParticleSwarms



#!/usr/bin/env python

import numpy as np
from olympus.objects import Object, Parameter
from olympus import Logger

from . import Object, Parameter

#===============================================================================

class ObjectParameterVector(Object):

	def __init__(self, array=None, dict=None, param_space=None):
		"""
        Args:
            array (array):
            dict (dict): dictionary with parameter names and values.
            param_space (ParameterSpace): ParameterSpace instance. This is typically defined as part of a Dataset and
                is also inherited by Emulator. If a `param_space` is defined, `info_dict` will be checked to ensure the
                provided keys match those in `param_space`, otherwise `info_dict` is accepted as is. Default is None.
        """
		from olympus.campaigns import ParameterSpace
		self.param_space = ParameterSpace()
		Object.__init__(self)

		if array is not None and param_space is not None:
			_ = self.from_array(array=array, param_space=param_space)
		elif dict is not None:
			_ = self.from_dict(info_dict=dict, param_space=param_space)

	def __repr__(self):
		string = 'ParamVector('
		for param in self.param_space:
			string += '{} = {}, '.format(param.name, getattr(self, param.name))
		if len(string) > 12:
			return string[:-2] + ')'
		else:
			return string + ')'

	def __str__(self):
		return self.__repr__()

	#***************************************************************************

	def __abs__(self):
		return np.abs(self.to_array())

	def __gt__(self, value):
		array = self.to_array()
		if array.shape[0] > 1:
			Logger.log(f'undefined operation: {self} > {value}', 'ERROR')
			raise SystemExit
		return array[0] > value

	def __lt__(self, value):
		array = self.to_array()
		if array.shape[0] > 1:
			Logger.log(f'undefined operation: {self} < {value}', 'ERROR')
			raise SystemExit
		return array[0] < value

	def __sub__(self, vector):
		array = self.to_array()
		if array.shape[0] > 1:
			Logger.log(f'undefined operation: {self} < {value}', 'ERROR')
			raise SystemExit
		return ObjectParameterVector().from_array(array - vector, self.param_space)

	#***************************************************************************

	def from_array(self, array, param_space):
		self.param_space = param_space
		for value, param in zip(array, param_space):
			self.add(param.name, value)
		return self

	def to_array(self):
		array = np.array([getattr(self, param_name) for param_name in self.param_space.param_names])
		return array

	def from_dict(self, info_dict, param_space=None):
		""" Creates a ParamVector representation of a given dictionary.

        Args:
            info_dict (dict): dictionary with parameter names and values.
            param_space (ParameterSpace): ParameterSpace instance. This is
				typically defined as part of a Dataset and is also inherited by
				Emulator. If a `param_space` is defined, `info_dict` will be
				checked to ensure the provided keys match those in `param_space`,
				otherwise `info_dict` is accepted as is. Default is None.
        """

		if param_space is None:
			for key, value in info_dict.items():
				# define parameter of parameter space
				self.param_space.add(Parameter(name=key))
				# add specific value for the parameter
				self.add(key, value)

		elif param_space is not None:
			if set(param_space.param_names) != set(list(info_dict.keys())):
				message = 'The dictionary keys provided do not match those in the parameter space'
				Logger.log(message, 'ERROR')
			self.param_space = param_space
			for param_name in param_space.param_names:
				# add specific value for the parameter
				self.add(param_name, info_dict[param_name])
		return self



#!/usr/bin/env python

from olympus.surfaces import GaussianMixture
from olympus import Logger


class Denali(GaussianMixture):

    def __init__(self, noise=None):
        """Gaussian Mixture surface derived from the GaussianMixture generator using ``random_seed=6190``.

        Args:
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        peak = 6190
        GaussianMixture.__init__(self, param_dim=2, num_gauss=3, cov_scale=0.1, diagonal_cov=False,
                                 noise=noise, random_seed=peak)

    @property
    def minima(self):
        # TODO: find minimum numerically and hardcode
        message = 'Unknown minima: these need to be found numerically'
        Logger.log(message, 'WARNING')
        return None

    @property
    def maxima(self):
        # TODO: find minimum numerically and hardcode
        message = 'Unknown maxima: these need to be found numerically'
        Logger.log(message, 'WARNING')
        return None





#!/usr/bin/env python

import time
import numpy as np

from olympus.objects  import ParameterVector
from olympus.planners import AbstractPlanner
from olympus.planners.utils_planner import get_init_guess, get_bounds
from olympus.utils    import daemon

#===============================================================================

class Snobfit(AbstractPlanner):

    SUBMITTED_PARAMS = []
    RECEIVED_VALUES  = []

    def __init__(self, goal='minimize',
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        '''
        Args:
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        '''
        AbstractPlanner.__init__(**locals())
        self.has_optimizer = False
        self.is_converged  = False
        self.budget        = 10**8

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.bounds      = get_bounds(param_space)
        if self.init_guess is None:
            self.init_guess = get_init_guess(param_space, method=self.init_guess_method, random_seed=self.init_guess_seed)

    def _tell(self, observations):
        self._params = observations.get_params(as_array = False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, param):
        self.SUBMITTED_PARAMS.append(param)
        while len(self.RECEIVED_VALUES) == 0:
            time.sleep(0.1)
        measurement = self.RECEIVED_VALUES.pop(0)
        return np.squeeze(measurement)

    @daemon
    def create_optimizer(self):
        from SQSnobFit import minimize, optset
        options = optset(maxmp=2+6)
        _, __ = minimize(self._priv_evaluator, x0=self.init_guess, bounds=self.bounds, budget=self.budget)#, options=options)
        self.is_converged = True


    def _ask(self):
        if self.has_optimizer is False:
            self.create_optimizer()
            self.has_optimizer = True

        while len(self.SUBMITTED_PARAMS) == 0:
            time.sleep(0.1)
            if self.is_converged:
                return ParameterVector().from_dict(self._params[-1])
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector().from_array(params, self.param_space)



#!/usr/bin/env python

# simple interface to the available scripts
from argparse import ArgumentParser, RawTextHelpFormatter, SUPPRESS
from olympus import __version__
import sys


class OlympusCli:
    def __init__(self):
        self.parser = ArgumentParser(
            description="""
    ------------------------------
    Olympus command line interface
    ------------------------------
    Available commands are:
        baseline     download baseline results with random search
        download     downloads dataset from olympus_datasets
        upload       uploads dataset to olympus_datasets
        etc          Etc""",
            formatter_class=RawTextHelpFormatter,
        )

        self.parser.add_argument(
            "-v", "--version", action="version", version=__version__
        )

        self._register_subparsers()
        self._parse_args()

    def _register_subparsers(self):
        self.subparsers = self.parser.add_subparsers(
            help="subparser help", dest="subparser_name"
        )

        from .cli_baseline import ParserBaseline
        from .cli_download import ParserDownload
        from .cli_upload import ParserUpload

        self.parser_baseline = ParserBaseline(self.subparsers)
        self.parser_download = ParserDownload(self.subparsers)
        self.parser_upload = ParserUpload(self.subparsers)

    def _parse_args(self):

        # first parse
        args, rest = self.parser.parse_known_args()
        if args.subparser_name is None:
            self.parser.print_help()
            return

        subparser = getattr(self, f"parser_{args.subparser_name}")
        success = subparser(args)
        # subsequent parse
        while len(rest) > 0 and success:
            args, rest = self.parser.parse_known_args(rest)
            subparser = getattr(self, f"parser_{args.subparser_name}")
            success = subparser(args)


def check_unknown_cmd(unknowns):
    """Checks unknown command line arguments are raises a warning if unexpected
	commands are found.
	"""
    expected = ["olympus", "get"]

    for cmd in unknowns:
        if cmd not in expected:
            print(
                'Unknown command found in your command line: "{}". '
                "This command will be ignored".format(cmd)
            )


def entry_point():
    OlympusCli()


if __name__ == "__main__":
    entry_point()



#!/usr/bin/env python

import numpy as np
from scipy.stats import gamma
from olympus.noises import AbstractNoise


class GammaNoise(AbstractNoise):

    def __init__(self, scale=1, lower_bound=0):
        """Gamma-distributed noise parametrised by its standard deviation (``scale``) and its mode.

        Args:
            scale (float): Standard deviation of the Gaussian. Default is 1.
            lower_bound (float): Lower bound of the Gamma distribution. The distribution will not have support below this
                lower bound. Default is 0.
        """
        AbstractNoise.__init__(**locals())

    def _add_noise(self, value):
        # value = mode
        value = value - self.lower_bound
        var = self.scale ** 2.
        s = np.sqrt(var + (value ** 2.) / 4.) - value / 2.
        k = value / s + 1.
        noisy_value = gamma.rvs(a=k, loc=self.lower_bound, scale=s)
        return noisy_value



#!/usr/bin/env python

#===============================================================================

import time

from olympus.objects                import ParameterVector
from olympus.planners               import AbstractPlanner
from olympus.planners.utils_planner import get_init_guess, get_bounds
from olympus.utils                  import daemon

#===============================================================================

class WrapperScipy(AbstractPlanner):

    KNOWS_BOUNDS     = False

    def __init__(self, *args, **kwargs):
        AbstractPlanner.__init__(self, *args, **kwargs)
        self.has_minimizer = False
        self.is_converged  = False

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.bounds      = get_bounds(param_space)
        if self.init_guess is None:
            self.init_guess = get_init_guess(param_space, method=self.init_guess_method, random_seed=self.init_guess_seed)


    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, params):
        if self.KNOWS_BOUNDS is False:
            params = self._project_into_domain(params)
        self.SUBMITTED_PARAMS.append(params)
        while len(self.RECEIVED_VALUES) == 0:
            time.sleep(0.1)
        value = self.RECEIVED_VALUES.pop(0)
        return value

    @daemon
    def create_minimizer(self):
        from scipy.optimize import minimize
        config = self.config.to_dict().copy()
        for attr in ['init_guess', 'init_guess_seed', 'init_guess_method']:
            if attr in config: del config[attr]
        if self.KNOWS_BOUNDS:
            _ = minimize(self._priv_evaluator, x0=self.init_guess, bounds=self.bounds, method=self.METHOD, options=config)
        else:
            _ = minimize(self._priv_evaluator, x0=self.init_guess, method=self.METHOD, options=config)
        self.is_converged = True

    def _ask(self):
        if self.has_minimizer is False:
            self.create_minimizer()
            self.has_minimizer = True

        while len(self.SUBMITTED_PARAMS) == 0:
            time.sleep(0.1)
            if self.is_converged:
                return ParameterVector(dict=self._params[-1])
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector(array=params, param_space=self.param_space)



#!/usr/bin/env python

#==============================================================================

import copy
import time
import sqlalchemy as sql

#==============================================================================

from olympus.campaigns                 import Campaign
from olympus.databases.database_sqlite import AddEntry, FetchEntries, UpdateEntries
from olympus.utils                     import daemon, thread, generate_id

#==============================================================================

class SqliteInterface:

	SQLITE_COLUMNS = {
		'dict':   sql.PickleType(),
		'float':  sql.Float(),
		'int':    sql.Integer(),
		'pickle': sql.PickleType(),
		'string': sql.String(512),
	}

	def __init__(self, name, path, attributes, verbosity = 0):
		self.TASKS_READ   = {}
		self.TASKS_UPDATE = []
		self.TASKS_WRITE  = []

		self.attributes = attributes
		self.db_path    = f'sqlite:///{path}'
		self.name       = name

		self._create_database()
		self._process_requests()


	def _create_database(self):
		# create database
		self.db       = sql.create_engine(self.db_path)
		self.db.echo  = False
		self.metadata = sql.MetaData(self.db)

		# create table in database
		self.table = sql.Table(self.name, self.metadata)
		for attr in self.attributes:
			attr_type = 'pickle'
			if attr['type'] in self.SQLITE_COLUMNS:
				attr_type = attr['type']
			self.table.append_column(sql.Column(attr['name'].lower(), self.SQLITE_COLUMNS[attr_type]))
		self.table.create(checkfirst = True)

	#*****************************************************************************

	@thread
	def _process_requests(self):
		self.is_processing = True
		keep_processing    = True
		iter_index         = 0
		while keep_processing:
			time.sleep(0.01)
			num_tasks_read   = len(self.TASKS_READ)
			num_tasks_update = len(self.TASKS_UPDATE)
			num_tasks_write  = len(self.TASKS_WRITE)
			iter_index += 1

			# run at most one reading request
			read_keys = copy.deepcopy(list(self.TASKS_READ.keys()))
			for read_key in read_keys:
				if not self.TASKS_READ[read_key].executed:
					self.TASKS_READ[read_key].execute()
					break
			else:
				num_tasks_read = 0

			# run all update requests
			for update_index in range(num_tasks_update):
				task_update = self.TASKS_UPDATE.pop(0)
				task_update.execute()

			# run at most one writing request
			if num_tasks_write > 0:
				task_write = self.TASKS_WRITE.pop(0)
				task_write.execute()

			# clean reading reqeusts
			delete_keys = []
			for read_key in read_keys:
				if not read_key in self.TASKS_READ: continue
				if self.TASKS_READ[read_key].entries_fetched:
					delete_keys.append(read_key)
			for read_key in delete_keys:
				del self.TASKS_READ[read_key]

#			keep_processing = len(self.TASKS_READ) > 0 or len(self.TASKS_UPDATE) > 0 or len(self.TASKS_WRITE) > 0
			keep_processing = num_tasks_read > 0 or num_tasks_update > 0 or num_tasks_write > 0
		self.is_processing = False

	#*****************************************************************************

	def return_dict(function):
		def wrapper(self, *args, **kwargs):
			entries    = function(self, *args, **kwargs)
			info_dicts = [Campaign().from_dict({key['name'].lower(): entry[key['name'].lower()] for key in self.attributes}) for entry in entries]
			return info_dicts
		return wrapper

	#*****************************************************************************

	def add(self, info_dict):
		add_entry = AddEntry(self.db, self.table, info_dict)
		self.TASKS_WRITE.append(add_entry)
		if not self.is_processing:
			self._process_requests()

	@return_dict
	def fetch_all(self, cond_dict = {}):
		cond_keys = list(cond_dict)
		cond_vals = list(cond_dict)

		# define selection
		selection = sql.select([self.table])
		for index, key in enumerate(cond_keys):
			if isinstance(cond_vals[index], list):
				# with a list, we need to combine all possibilities with _or
				if len(cond_vals[index]) == 0:
					return []
				filters   = [getattr(self.table.c, key) == value for value in cond_vals[index]]
				condition = sql.or_(*filters)
			else:
				condition = getattr(self.table.c, key) == cond_vals[index]
			selection = selection.where(condition)

		# create fetch request
		fetch_entries = FetchEntries(self.db, self.table, selection, name = self.name)
		fetch_keys    = generate_id()
		self.TASKS_READ[fetch_keys] = fetch_entries
		if not self.is_processing:
			self._process_requests()

		entries = fetch_entries.get_entries()
		return entries


	def update_all(self, cond_dict, update_dict):
		cond_keys = list(cond_dict.keys())
		cond_vals = list(cond_dict.values())

		# define selection
		update = sql.update(self.table).values(update_dict).where(getattr(self.table.c, cond_keys[0]) == cond_vals[0])
		for index, key in enumerate(cond_keys[1:]):
			update = update.where(getattr(self.table.c, key) == cond_vals[index + 1])

		# submitting update
		update_entries = UpdateEntries(self.db, self.table, update)
		self.TASKS_UPDATE.append(update_entries)
		if not self.is_processing:
			self._process_requests()



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface
from itertools import product
from olympus import Logger


class DiscreteDoubleWell(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Discrete double well function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        params = [0.16451863] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        message = 'DiscreteDoubleWell has an infinite number of maxima'
        Logger.log(message, 'INFO')
        # some maxima
        maxima = []
        params = product([0, 1], repeat=self.param_dim)
        for param in params:
            param = list(param)
            value = self._run(param)
            maxima.append({'params': param, 'value': value})
        return maxima

    def _run(self, params):
        params = np.array(params)
        params = 100 * params - 15

        bounds = [1.25, 1.75, 2.0, 2.5, 2.75]
        bounds = np.array(bounds) ** 2
        result = 5
        for bound in bounds[::-1]:
            if np.amax(np.abs(params)) < bound:
                result -= 1

        params -= 50
        params[0] += 10
        bounds = [2.5, 4.0, 5.0, 6.5]
        bounds = np.array(bounds) ** 2
        new_res = 5
        for bound in bounds[::-1]:
            if np.amax(np.abs(params)) < bound:
                new_res -= 1

        result = np.amin([result, 4, new_res])

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

import numpy as np

from olympus.planners.wrapper_scipy import WrapperScipy
from olympus.objects.object_config  import Config


class ConjugateGradient(WrapperScipy):

    METHOD = 'cg'

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy/reference/optimize.minimize-neldermead.html#optimize-minimize-neldermead
    def __init__(self, goal='minimize', disp=False, maxiter=None, gtol=1e-05, norm=np.inf, eps=1.4901161193847656e-8,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """
        Conjugate Gradient optimizer based on the SciPy implementation.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            disp (bool): Set to True to print convergence messages.
            maxiter (int): Maximum number of iterations to perform.
            gtol (float): Gradient norm must be less than gtol before successful termination.
            norm (float): Order of norm (Inf is max, -Inf is min).
            eps (float or ndarray): If jac is approximated, use this value for the step size.
            init_guess (array, optional): initial guess for the optimization
            init_guess_method (str): method to construct initial guesses if init_guess is not provided.
                Choose from: random
            init_guess_seed (str): random seed for init_guess_method
        """

        WrapperScipy.__init__(**locals())


#===============================================================================

if __name__ == '__main__':

    from olympus import Parameter, ParameterSpace
    param_space = ParameterSpace()
    param_space.add(Parameter(name='param_0'))
    param_space.add(Parameter(name='param_1'))

    planner = ConjugateGradient()
    planner.set_param_space(param_space=param_space)
    param = planner.ask()
    print('PARAM', param)



#!/usr/bin/env python

#===============================================================================

import uuid
import traceback

#===============================================================================

def generate_id():
    identifier = str(uuid.uuid4())[:8]
    return identifier

#===============================================================================

def check_module(module_name, message, **kwargs):
    try:
        _ = __import__(module_name)
    except ModuleNotFoundError:
        from olympus import Logger
        error = traceback.format_exc()
        for line in error.split('\n'):
            if 'ModuleNotFoundError' in line:
                module = line.strip().strip("'").split("'")[-1]
        kwargs.update(locals())
        message = f'{message}'.format(**kwargs)
        Logger.log(message, 'ERROR')

def check_planner_module(planner, module, link):
    from olympus.planners import get_planners_list
    message = '''Planner {planner} requires {module} which could not be found. Please install
    {module} and check out {link} for further instructions or use another
    available planner: {planners_list}'''
    check_module(module, message, planner=planner, planners_list=', '.join(get_planners_list()), link=link)

#===============================================================================



"""Olympus
"""

from setuptools import setup
from setuptools import find_packages

import versioneer


def readme():
    with open("README.md") as f:
        return f.read()


# taken from https://hanxiao.io/2019/11/07/A-Better-Practice-for-Managing-extras-require-Dependencies-in-Python/
def get_extra_requires(path, add_all=True):
    import re
    from collections import defaultdict

    with open(path) as content:
        extra_deps = defaultdict(set)
        for entry in content:
            if entry.strip() and not entry.startswith("#"):
                tags = set()
                if ":" in entry:
                    package, tag_list = entry.split(":")
                    tags.update(tag_entry.strip() for tag_entry in tag_list.split(","))
                tags.add(re.split("[<=>]", package)[0])
                for tag in tags:
                    extra_deps[tag].add(package)

        if add_all:
            extra_deps["all"] = set(vv for v in extra_deps.values() for vv in v)
    return extra_deps


# -----
# Setup
# -----
setup(
    name="olymp",
    version="0.0.1b0",  #versioneer.get_version(),
#    cmdclass=versioneer.get_cmdclass(),
    description="Benchmarking framework for noisy optimization and experiment planning",
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
    ],
    url="https://github.com/aspuru-guzik-group/olympus",
    author="Florian Hase, Matteo Aldeghi, Riley Hickman",
    packages=find_packages('./src'),
    package_dir={"": "src"},
    package_data={"": ["datasets/*"]},
    include_package_data=True,
    zip_safe=False,
    tests_require=["pytest"],
    install_requires=["numpy", "pandas"],
    python_requires=">=3.6",
    extras_require=get_extra_requires("extra_requirements.txt"),
    entry_points={"console_scripts": ["olympus = olympus.cli.main:entry_point"]},
)



#!/usr/bin/env python

import os
import subprocess

# ===============================================================================

from olympus import __home__
from olympus import Logger

# ===============================================================================

try:
    import requests
except ModuleNotFoundError:
    Logger.log("module <requests> required to download dataset", "FATAL")

# ===============================================================================


class ConnectorServer:

    URL = "http://olympus.datasets.ngrok.io"

    def __init__(self):
        pass

    def list(self):
        Logger.log("connecting to server", "INFO")
        url = f"{self.URL}/list_datasets"
        print("URL", url)
        response = requests.post(url, data={})
        if response.status_code == 200:
            datasets = response.json()["datasets"]
            return sorted(datasets)
        else:
            return self._process_response(response)

        return self._process_response(response)

    def get_dataset(self, dataset_name):
        success = self._check_dataset(dataset_name)
        if success:
            Logger.log("downloading dataset", "INFO")
            self._download_dataset(dataset_name)

    def get_baseline(self):
        Logger.log("downloading baseline", "INFO")
        self._download_baseline()

    def _process_response(self, response):
        if response.status_code == 200:
            return True
        elif response.status_code == 404:
            Logger.log("could not reach server", "ERROR")
            return False
        else:
            Logger.log("unknown error", "ERROR")
            return False

    def _check_dataset(self, dataset_name):
        Logger.log("connecting to server", "INFO")
        url = f"{self.URL}/check_dataset"
        data = {"dataset_name": dataset_name}
        response = requests.post(url, data=data)
        return self._process_response(response)

    def _download_dataset(self, dataset_name):
        Logger.log(f"downloading dataset {dataset_name}", "INFO")
        url = f"{self.URL}/get_dataset"
        data = {"dataset_name": dataset_name}
        response = requests.post(url, data=data)
        if response.status_code == 200:

            target_dir = f"{__home__}/datasets/dataset_{dataset_name}"
            target_name = f"{target_dir}/dataset.zip"

            try:
                os.makedirs(target_dir)
            except:
                pass

            # save file
            Logger.log("saving dataset", "INFO")
            with open(target_name, "wb") as content:
                content.write(response.content)

            # unzip file
            Logger.log("unpacking dataset", "INFO")
            subprocess.call(f"unzip {target_name} -d {target_dir}", shell=True)
            Logger.log("dataset installed", "INFO")

        elif response.status_code == 204:
            error = response.json()["error"]
            Logger.log(f"did not get dataset ({error})", "ERROR")

    def _download_baseline(self):
        url = f"{self.URL}/get_baseline"
        response = requests.post(url, data={})
        if response.status_code == 200:

            target_dir = f"{__home__}/baseline/"
            target_name = f"{target_dir}/baseline.zip"

            if not os.path.isdir(target_dir):
                os.makedirs(target_dir)

            Logger.log("saving baseline", "INFO")
            with open(target_name, "wb") as content:
                content.write(response.content)

            # unzip baseline
            Logger.log("unpacking baseline", "INFO")
            subprocess.call(f"unzip {target_name} -d {target_dir}", shell=True)
            Logger.log("baseline installed", "INFO")
            os.remove(target_name)

        elif response.status_code == 204:
            error = response.json()["error"]
            Logger.log(f"did not get baseline ({error})", "ERROR")



#!/usr/bin/env python

import functools
import threading

#===============================================================================

def thread(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target = function, args = args, kwargs = kwargs)
        thread.start()
    return wrapper

def daemon(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target = function, args = args, kwargs = kwargs)
        thread.daemon = True
        thread.start()
    return wrapper
    
#===============================================================================



#!/usr/bin/env python

import numpy as np
from olympus.surfaces import AbstractSurface


class Zakharov(AbstractSurface):

    def __init__(self, param_dim=2, noise=None):
        """Zakharov function.

        Args:
            param_dim (int): Number of input dimensions. Default is 2.
            noise (Noise): Noise object that injects noise into the evaluations of the surface. Default is None.
        """
        AbstractSurface.__init__(**locals())

    @property
    def minima(self):
        params = [1./3.] * self.param_dim
        value = self._run(params)
        return [{'params': params, 'value': value}]

    @property
    def maxima(self):
        return None

    def _run(self, params):
        params = np.array(params)
        params = 15 * params - 5  # rescale onto [-5, 10]

        term1 = np.sum(params**2)
        term2 = 0.
        for i, x in enumerate(params):
            term2 += 0.5 * (i+1) * x
        term3 = term2**4
        term2 = term2**2

        result = term1 + term2 + term3

        if self.noise is None:
            return result
        else:
            return self.noise(result)



#!/usr/bin/env python

import os
import time
import numpy as np
import silence_tensorflow

silence_tensorflow.silence_tensorflow()
import tensorflow as tf
from sklearn.metrics import r2_score, mean_squared_error

from olympus import Logger
from olympus.models import AbstractModel

# ===============================================================================

os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

# ===============================================================================


class WrapperTensorflowModel(AbstractModel):
    def __init__(self, *args, **kwargs):
        AbstractModel.__init__(self, *args, **kwargs)
        self.graph = None
        self.is_graph_constructed = False
        tf.compat.v1.reset_default_graph()

    def act_funcs(self, act_func_name):
        activation_functions = {
            "linear": lambda y: y,
            "leaky_relu": lambda y: tf.nn.leaky_relu(y, 0.2),
            "relu": lambda y: tf.nn.relu(y),
            "softmax": lambda y: tf.nn.softmax(y),
            "softplus": lambda y: tf.nn.softplus(y),
            "softsign": lambda y: tf.nn.softsign(y),
            "sigmoid": lambda y: tf.nn.sigmoid(y),
        }
        return activation_functions[act_func_name]

    def _generator(self, features, targets):
        batch_indices = np.random.randint(
            low=0, high=features.shape[0], size=self.batch_size
        )
        features_batch = features[batch_indices]
        targets_batch = targets[batch_indices]
        return features_batch, targets_batch

    def _set_dims(self, features_dim, targets_dim):
        self.features_dim = features_dim
        self.targets_dim = targets_dim

    # def _project_features(self, features, transformer):
    #    projections = []
    #    transformed_features = transformer.back_transform(features)
    #    for _, param_definition in enumerate(self.param_space):
    #        feature_values = features[:, _]
    #        if param_definition.is_periodic is True:
    #            # compute sine and cosine
    #            cosine = np.cos( 2 * np.pi * (transformed_features[:, _] - param_definition.low) / (param_definition.high - param_definition.low))
    #            sine   = np.sin( 2 * np.pi * (transformed_features[:, _] - param_definition.low) / (param_definition.high - param_definition.low))
    #            projections.append(cosine)
    #            projections.append(sine)
    #        elif param_definition.is_periodic is False:
    #            projections.append(feature_values)
    #   #             projections.append(transformed_features[:, _])
    #        else:
    #            raise NotImplementedError
    #    projections = np.array(projections).T
    #    return projections

    # def train(self, train_features, train_targets, valid_features, valid_targets, model_path,
    #          plot=False, feature_transformer=None, target_transformer=None):
    def train(
        self,
        train_features,
        train_targets,
        valid_features,
        valid_targets,
        model_path,
        plot=False,
    ):
        """

        Args:
            train_features: features of the training set.
            train_targets: targets of the training set.
            valid_features: features of the validation/test set, used for early stopping.
            valid_targets: targets of the validation/test set, used for early stopping.
            model_path (str): where the TensorFlow checkpoints will be written.
            plot (bool): whether to show scatter plots of the training progression.
            feature_transformer:
            target_transformer:

        Returns:
            max_train_r2: best r2 score on training set
            max_valid_r2: best r2 score on validation set
        """
        # train_features = self._project_features(train_features, feature_transformer)
        # valid_features = self._project_features(valid_features, feature_transformer)

        # compute needed stats for features and targets
        self._set_dims(
            features_dim=np.shape(train_features)[1],
            targets_dim=np.shape(train_targets)[1],
        )
        if not self.is_graph_constructed:
            self._build_inference()
        train_errors, valid_errors, losses = list(), list(), list()

        start_time = time.time()
        with self.graph.as_default():
            with self.sess.as_default():
                self.saver = tf.compat.v1.train.Saver()

                if plot:
                    import matplotlib.pyplot as plt

                    fig = plt.figure(figsize=(12, 6))
                    ax0 = plt.subplot2grid((1, 2), (0, 0))
                    ax1 = plt.subplot2grid((1, 2), (0, 1))
                    axs = [ax0, ax1]
                    plt.ion()

                if len(valid_targets) > self.batch_size:
                    valid_indices = np.random.randint(
                        low=0, high=len(valid_targets), size=self.batch_size
                    )
                else:
                    valid_indices = np.arange(len(valid_targets))

                if len(train_targets) > self.batch_size:
                    train_indices = np.random.randint(
                        low=0, high=len(train_targets), size=self.batch_size
                    )
                else:
                    train_indices = np.arange(len(train_targets))

                # print out info
                _print_header()
                for epoch in range(self.max_epochs):

                    train_features_batch, train_targets_batch = self._generator(
                        train_features, train_targets
                    )
                    __, loss = self.sess.run(
                        [self.train_op, self.loss],
                        feed_dict={
                            self.tf_x: train_features_batch,
                            self.tf_y: train_targets_batch,
                        },
                    )
                    losses.append(loss)

                    if epoch % self.pred_int == 0:

                        # make a prediction on the validation set
                        valid_pred = self.predict(
                            features=valid_features[valid_indices], num_samples=10
                        )
                        valid_r2 = r2_score(valid_targets[valid_indices], valid_pred)
                        valid_rmsd = np.sqrt(
                            mean_squared_error(valid_targets[valid_indices], valid_pred)
                        )
                        valid_errors.append([valid_r2, valid_rmsd])

                        # make a prediction on the train set
                        train_pred = self.predict(
                            features=train_features[train_indices], num_samples=10
                        )
                        train_r2 = r2_score(train_targets[train_indices], train_pred)
                        train_rmsd = np.sqrt(
                            mean_squared_error(train_targets[train_indices], train_pred)
                        )
                        train_errors.append([train_r2, train_rmsd])

                        if plot:
                            for ax in axs:
                                ax.cla()

                            plot_train_targets = train_targets[train_indices]
                            plot_train_pred = train_pred
                            plot_valid_targets = valid_targets[valid_indices]
                            plot_valid_pred = valid_pred

                            plot_min_targets, plot_max_targets = (
                                np.amin(plot_train_targets),
                                np.amax(plot_train_targets),
                            )
                            ax0.plot(
                                [plot_min_targets, plot_max_targets],
                                [plot_min_targets, plot_max_targets],
                                color="#444444",
                            )
                            ax0.plot(
                                plot_train_targets, plot_train_pred, marker="o", ls=""
                            )
                            ax0.plot(
                                plot_valid_targets, plot_valid_pred, marker="o", ls=""
                            )
                            ax1.plot(np.array(train_errors)[:, 0])
                            ax1.plot(np.array(valid_errors)[:, 0])
                            plt.pause(0.05)

                        min_rmsd_index = np.argmin(np.array(valid_errors)[:, 1])
                        if len(valid_errors) - min_rmsd_index > self.es_patience:
                            break

                        newline = f"{epoch:>15}{train_r2:>15.3f}{train_rmsd:>15.3f}{valid_r2:>15.3f}{valid_rmsd:>15.3f}"
                        # the latest model is the best ==> save it and tag it on screen
                        if min_rmsd_index == len(valid_errors) - 1:
                            self.saver.save(self.sess, f"{model_path}/model.ckpt")
                            newline += " *"
                        Logger.log(newline, "INFO")

                # report the train and valid performance of the model we saved
                # Note we saved the best model based on the lowest RMSE
                mdl_train_r2 = np.array(train_errors)[min_rmsd_index, 0]
                mdl_valid_r2 = np.array(valid_errors)[min_rmsd_index, 0]
                mdl_train_rmsd = np.array(train_errors)[min_rmsd_index, 1]
                mdl_valid_rmsd = np.array(valid_errors)[min_rmsd_index, 1]

        Logger.log(
            f"Training completed in {round(time.time() - start_time, 2)} seconds.",
            "INFO",
        )
        Logger.log("=" * 75 + "\n", "INFO")
        return mdl_train_r2, mdl_valid_r2, mdl_train_rmsd, mdl_valid_rmsd

    def restore(self, model_path):
        if not self.is_graph_constructed:
            tf.compat.v1.reset_default_graph()
            self._build_inference()
        with self.graph.as_default():
            with self.sess.as_default():
                try:
                    self.saver.restore(self.sess, model_path + "/model.ckpt")
                    return True
                # TODO: I get the error 'tf has no attribute python' here
                except tf.python.framework.errors_impl.InvalidArgumentError:
                    return False

    def predict(self, features, num_samples=1):
        # features = self._project_features(features, feature_transformer)
        # make sure the dimensionality of the input matches that used for training
        if features.shape[1] != self.features_dim:
            raise ValueError(
                "dimensionality of input features provided does not match that of the training dataset"
            )

        with self.sess.as_default():
            pred = np.empty((num_samples, len(features), 1))
            resolution = divmod(len(features), self.batch_size)
            res = [self.batch_size for i in range(resolution[0])]
            res.append(resolution[1])
            res = list(filter((0).__ne__, res))
            res_cu = [i * self.batch_size for i in range(len(res))]
            res_stp = [res_cu[_] + re for _, re in enumerate(res)]

            for batch_iter, size in enumerate(res):
                start, stop = res_cu[batch_iter], res_stp[batch_iter]
                X_test_batch = None
                if size == self.batch_size:
                    X_test_batch = features[start:stop]
                elif size != self.batch_size:
                    X_test_batch = np.concatenate(
                        (
                            features[start:stop],
                            np.random.choice(
                                features[:, 0],
                                size=(self.batch_size - size, features.shape[1]),
                            ),
                        ),
                        axis=0,
                    )

                for _ in range(num_samples):
                    predic = self.sess.run(
                        self.y_pred, feed_dict={self.tf_x: X_test_batch}
                    )
                    pred[_, start:stop] = predic[:size]

            pred = np.mean(pred, axis=0)
            return pred  # return the scaled prediction (Emulator will back-transform)


def _print_header():
    Logger.log(
        "    =======================================================================",
        "INFO",
    )
    Logger.log(
        "{0:>15}{1:>15}{2:>15}{3:>15}{4:>15}".format(
            "Epoch", "Train R2", "Train RMSD", "Test R2", "Test RMSD"
        ),
        "INFO",
    )
    Logger.log(
        "    =======================================================================",
        "INFO",
    )



#!/usr/bin/env python

import numpy as np
from olympus.emulators.emulator import Emulator
from olympus.surfaces.surface import AbstractSurface
from olympus.objects import Object
from olympus.utils import generate_id
from olympus.campaigns import ParameterSpace
from olympus.campaigns import Observations


class Campaign(Object):
    """ stores information about a single optimization run

    This class logs all information about a single optimization run,

    """

    ATT_ACCEPTS = {
        "type": "string",
        "default": "param_vector",
        "valid": ["param_vector", "array", "dict"],
    }
    ATT_GOAL = {
        "type": "string",
        "default": "minimize",
        "valid": ["minimize", "maximize"],
    }
    ATT_EMULATOR_TYPE = {
        "type": "string",
        "default": "n/a",
        "valid": ["numeric", "analytic"],
    }
    ATT_DATASET_KIND = {"type": "string", "default": "n/a"}
    ATT_MEASUREMENT_NAME = {"type": "string", "default": "n/a"}
    ATT_MODEL_KIND = {"type": "string", "default": "n/a"}
    ATT_SURFACE_KIND = {"type": "string", "default": "n/a"}
    ATT_ID = {"type": "string", "default": generate_id}
    ATT_OBSERVATIONS = {"type": "Observations", "default": Observations}
    ATT_PARAM_SPACE = {"type": "ParameterSpace", "default": ParameterSpace}
    ATT_PLANNER_KIND = {"type": "string", "default": "n/a"}

    def __repr__(self):
        repr_ = f"<Campaign (dataset={self.dataset_kind}, model={self.model_kind}, planner={self.planner_kind}, num_iter={len(self.params)})>"
        return repr_

    @property
    def params(self, *args, **kwargs):
        return self.observations.get_params(*args, **kwargs)

    @property
    def values(self, *args, **kwargs):
        return self.observations.get_values(*args, **kwargs)

    @property
    def best_values(self):
        vals = self.observations.get_values()
        if len(vals) == 0:
            return vals
        best_vals = [vals[0]]
        for val in vals[1:]:
            if self.goal == "minimize":
                best_vals.append(np.minimum(val, best_vals[-1]))
            elif self.goal == "maximize":
                best_vals.append(np.maximum(val, best_vals[-1]))
        best_vals = np.array(best_vals)
        return best_vals

    def add_observation(self, param, value):
        self.observations.add_observation(param, value)

    def set_planner_specs(self, planner):
        self.set_planner_kind(planner.kind)
        self.set_goal(planner.goal)

    def set_param_space(self, param_space):
        self.param_space = param_space
        self.observations.set_param_space(param_space)

    def set_value_space(self, value_space):
        self.value_space = value_space
        self.observations.set_value_space(value_space)

    def set_emulator_specs(self, emulator):
        """Store info about the Emulator (or Surface) into the campaign object.

        Args:
            emulator (object): an Emulator or Surface instance.

        """
        self.set_accepts("array")
        if isinstance(emulator, Emulator):
            self.set_emulator_type("numeric")
            self.set_dataset_kind(emulator.dataset.kind)
            self.set_measurement_name(emulator.dataset.measurement_name)
            self.set_model_kind(emulator.model.kind)
        elif isinstance(emulator, AbstractSurface):
            self.set_emulator_type("analytic")
            self.set_surface_kind(emulator.kind)

        self.set_param_space(emulator.param_space)
        self.set_value_space(emulator.value_space)

    # def set_surface_specs(self, surface):
    # 	self.set_accepts('array')
    # 	self.set_surface_name(surface.name)
    # self.set_goal(surface.goal)
    # 	self.set_param_space(surface.get_param_space())

    def to_dict(self):
        return Object.to_dict(self, exclude=["func"])


# ===============================================================================



#!/usr/bin/env python

import time
from olympus.objects  import ParameterVector
from olympus.planners import AbstractPlanner
from olympus.utils    import daemon
import numpy as np


class ParticleSwarms(AbstractPlanner):

    def __init__(self, goal='minimize', max_iters=10**8, options={'c1': 0.5, 'c2': 0.3, 'w': 0.9}, particles=10):
        """
        Particle swarm optimizer.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            max_iters (int): The maximum number of iterations for the swarm to search.
            options (dict): ???
            particles (int): The number of particles in the swarm.
        """
        AbstractPlanner.__init__(**locals())
        self.has_optimizer = False
        self.is_converged  = False

    def _set_param_space(self, param_space):
        self.param_space = param_space

    def _tell(self, observations):
        self._params = observations.get_params(as_array = False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, params_array):
        for params in params_array:
            params = self._project_into_domain(params)
            self.SUBMITTED_PARAMS.append(params)
        while len(self.RECEIVED_VALUES) < self.particles:
            time.sleep(0.1)
        measurements = np.array(self.RECEIVED_VALUES)
        self.RECEIVED_VALUES = []
        return measurements

    @daemon
    def create_optimizer(self):
        from pyswarms.single import GlobalBestPSO
        self.optimizer = GlobalBestPSO(
                n_particles=self.particles,
                options=self.options,
                dimensions=len(self.param_space))
        cost, pos = self.optimizer.optimize(self._priv_evaluator, iters=self.max_iters)
        self.is_converged = True

    def _ask(self):
        if self.has_optimizer is False:
            self.create_optimizer()
            self.has_optimizer = True

        while len(self.SUBMITTED_PARAMS) == 0:
            time.sleep(0.1)
            if self.is_converged:
                return ParameterVector().from_dict(self._params[-1])
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector().from_array(params, self.param_space)



#!/usr/bin/env python

# ======================================================================

import sys
import traceback

# ======================================================================


class MessageLogger:

    VERBOSITY_LEVELS = {
        -1: [],
        0: ["FATAL"],
        1: ["INFO", "FATAL"],
        2: ["INFO", "ERROR", "FATAL"],
        3: ["INFO", "WARNING", "ERROR", "FATAL"],
        4: ["DEBUG", "INFO", "WARNING", "ERROR", "FATAL"],
    }

    WRITER = {
        "DEBUG": sys.stdout,
        "INFO": sys.stdout,
        "WARNING": sys.stderr,
        "ERROR": sys.stderr,
        "FATAL": sys.stderr,
    }

    # more colors and styles:
    # https://stackoverflow.com/questions/2048509/how-to-echo-with-different-colors-in-the-windows-command-line
    # https://joshtronic.com/2013/09/02/how-to-use-colors-in-command-line-output/

    GREY = "0;37"
    WHITE = "1;37"
    YELLOW = ("1;33",)
    LIGHT_RED = ("1;31",)
    RED = "0;31"
    COLORS = {
        "DEBUG": WHITE,
        "INFO": GREY,
        "WARNING": YELLOW,
        "ERROR": LIGHT_RED,
        "FATAL": RED,
    }

    def __init__(self, verbosity=4):
        self.set_verbosity(verbosity)
        for key in self.WRITER:
            setattr(self, "{}S".format(key), [])

    def set_verbosity(self, verbosity):
        self.verbosity = verbosity
        self.verbosity_levels = self.VERBOSITY_LEVELS[self.verbosity]

    def log(self, message, message_type, only_once=False):

        if only_once:
            logs = getattr(self, f"{message_type}S")
            if message in logs:
                return

        logs = getattr(self, "{}S".format(message_type))
        logs.append(message)
        setattr(self, "{}S".format(message_type), logs)

        if only_once:
            message += "\n    [This message will be shown only once]"

        if not message_type in self.verbosity_levels:
            return None
        color = self.COLORS[message_type]
        writer = self.WRITER[message_type]
        if message_type in ["WARNING", "ERROR", "FATAL"]:
            error = traceback.format_exc()
            if error != "NoneType: None\n":
                writer.write(error)
        uncolored = "[{message_type}] {message}\n".format(
            message_type=message_type, message=message
        )
        message = "\x1b[%sm" % (color) + uncolored + "\x1b[0m"
        writer.write(message)

        if message_type == "FATAL":
            raise SystemExit

    def purge(self):
        for message_type in self.WRITER.keys():
            setattr(self, "{}S".format(message_type), [])


# ======================================================================



#!/usr/bin/env python

import time
import numpy as np 

from sobol_seq import i4_sobol

from olympus.objects                   import ParameterVector
from olympus.planners.abstract_planner import AbstractPlanner

#===============================================================================

class Sobol(AbstractPlanner):    

    def __init__(self, goal='minimize', budget=None, seed=None):
        AbstractPlanner.__init__(**locals())
        if seed is None:
            self.seed = np.random.randint(low = 0, high = 1000)

    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.dim = len(self.param_space)

    def _tell(self, observations):
        # random search does not need observations
        pass

    def _ask(self):
        vector, self.seed = i4_sobol(self.dim, self.seed)
        for index, param in enumerate(self.param_space):
            vector[index] = (param.high - param.low) * vector[index] + param.low
        return ParameterVector().from_array(vector, self.param_space)






#!/usr/bin/env python

import os
from glob import glob

olympus_home = os.path.dirname(os.path.abspath(__file__))
olympus_scratch = os.path.join(os.path.dirname(os.path.abspath(__file__)), ".scratch")
__home__ = os.environ.get("OLYMPUS_HOME") or olympus_home
__scratch__ = os.environ.get("OLYMPUS_SCRATCH") or olympus_scratch
__emulator_path__ = os.environ.get("OLYMPUS_EMULATOR_PATH") or os.path.abspath(
    os.path.join(__home__, "emulators")
)

# ===============================================================================
# Make sure that home and scratch exist
for dir_name in [__home__, __scratch__]:
    if not os.path.isdir(dir_name):
        os.mkdir(dir_name)

# ===============================================================================

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

# ===============================================================================
# this is where we can read in environment variables to modify
# the behavior of the logger
from .utils import MessageLogger

Logger = MessageLogger()

# ===============================================================================

from .objects import Object
from .objects import Parameter
from .objects import ParameterContinuous
from .objects import ParameterVector

from .campaigns import Campaign
from .campaigns import Observations
from .campaigns import ParameterSpace

from .analyzer import Analyzer
from .databases import Database
from .datasets import Dataset
from .emulators import Emulator
from .models import Model
from .planners import Planner
from .plotter import Plotter
from .surfaces import Surface
from .noises import Noise

from .olympus import Olympus

# ===============================================================================

from .datasets import list_datasets
from .emulators import list_trained_emulators
from .planners import list_planners

# ===============================================================================

from .baseline import Baseline



#!/usr/bin/env python

import pickle
import argparse


parser = argparse.ArgumentParser()

parser.add_argument('-f',
                    dest='file',
                    type=str,
                    help='Pickle file')
parser.add_argument('-k',
                    dest='kind',
                    help='New dataset kind/name',
                    type=str)
args = parser.parse_args()

# load
with open(args.file, 'rb') as content:
    emulator = pickle.load(content)

# rename
emulator.dataset.kind = args.kind

# save
with open('new_emulator.pickle', 'wb') as content:
    emulator = pickle.dump(emulator, content)



#!/usr/bin/env python

#===============================================================================

import numpy as np
import matplotlib.pyplot as plt

from olympus.plotter import AbstractPlotter

#===============================================================================

class PlotterMatplotlib(AbstractPlotter):

    def _plot(self, emulators, planners, measurements, file_name = None, show = False):
        num_plots  = len(emulators)
        num_graphs = len(planners)
        fig = plt.figure(figsize = (6, 4 * num_plots))
        axs = []
        for plot_index in range(num_plots):
            ax = plt.subplot2grid((num_plots, 1), (plot_index, 0))
            axs.append(ax)

        for plot_index, emulator in enumerate(emulators):
            ax = axs[plot_index]
            for planner in planners:

                idxs, vals = [], []
                indices = measurements[emulator][planner]['idxs'].copy()
                values  = measurements[emulator][planner]['vals'].copy()
                while len(indices) > 1:
                    idxs.append(indices[0])
                    vals.append(values[0])
                    indices = indices[1:]
                    values  = values[1:]
                    if idxs[-1] > indices[0]:
                        vals = np.squeeze(vals)
                        ax.plot(idxs, vals)
                        idxs, vals = [], []
                vals = np.squeeze(vals)
                ax.plot(idxs, vals)
                idxs, vals = [], []

            ax.set_title(f'{emulator.capitalize()}')

            ax.set_xlabel('# evaluations')
            ax.set_ylabel('measurement')

        plt.tight_layout()

        if file_name is not None:
            fig.savefig(file_name, bbox_inches = 'tight')
        if show is True:
            plt.show()



#!/usr/bin/env python

from olympus.utils.misc import check_planner_module

# ----------------
# Import planner
# ----------------

planner = 'Hyperopt'
module  = 'hyperopt'
link    = 'http://hyperopt.github.io/hyperopt'
check_planner_module(planner, module, link)

# ------------------------
# Import Planner Wrapper
# ------------------------

from .wrapper_hyperopt import Hyperopt



#!/usr/bin/env python


import os
__home__ = os.path.dirname(os.path.abspath(__file__))

#===============================================================================

from .abstract_database import AbstractDatabase
from .database          import Database

import glob
db_types  = []
databases = {}
for dir_name in glob.glob('{}/database_*'.format(__home__)):
    dir_name = dir_name.split('/')[-1][9:]
    db = Database(kind=dir_name)
    db_types.extend(db.file_types)
    databases[dir_name.lower()] = db

#===============================================================================



#!/usr/bin/env python

import time

from olympus.objects.object_config  import Config
from olympus.objects                import ParameterVector
from olympus.planners               import AbstractPlanner
from olympus.planners.utils_planner import get_init_guess, get_bounds
from olympus.utils                  import daemon


#===============================================================================

class BasinHopping(AbstractPlanner):

    # defaults are copied from scipy documentation
    # --> https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.optimize.basinhopping.html
    def __init__(self, goal='minimize', niter=100, T=1.0, stepsize=0.5, minimizer_kwargs=None,
                 take_step=None, accept_test=None, callback=None, interval=50,
                 disp=False, niter_success=None, seed=None,
                 init_guess=None, init_guess_method='random', init_guess_seed=None):
        """
        Find the global minimum of a function using the basin-hopping algorithm
        Parameters
        ----------
        goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
        func : callable ``f(x, *args)``
            Function to be optimized.  ``args`` can be passed as an optional item
            in the dict ``minimizer_kwargs``
        x0 : ndarray
            Initial guess.
        niter : integer, optional
            The number of basin hopping iterations
        T : float, optional
            The "temperature" parameter for the accept or reject criterion.  Higher
            "temperatures" mean that larger jumps in function value will be
            accepted.  For best results ``T`` should be comparable to the
            separation
            (in function value) between local minima.
        stepsize : float, optional
            initial step size for use in the random displacement.
        minimizer_kwargs : dict, optional
            Extra keyword arguments to be passed to the minimizer
            ``scipy.optimize.minimize()`` Some important options could be:
                method : str
                    The minimization method (e.g. ``"L-BFGS-B"``)
                args : tuple
                    Extra arguments passed to the objective function (``func``) and
                    its derivatives (Jacobian, Hessian).
        take_step : callable ``take_step(x)``, optional
            Replace the default step taking routine with this routine.  The default
            step taking routine is a random displacement of the coordinates, but
            other step taking algorithms may be better for some systems.
            ``take_step`` can optionally have the attribute ``take_step.stepsize``.
            If this attribute exists, then ``basinhopping`` will adjust
            ``take_step.stepsize`` in order to try to optimize the global minimum
            search.
        accept_test : callable, ``accept_test(f_new=f_new, x_new=x_new, f_old=fold, x_old=x_old)``, optional
            Define a test which will be used to judge whether or not to accept the
            step.  This will be used in addition to the Metropolis test based on
            "temperature" ``T``.  The acceptable return values are True,
            False, or ``"force accept"``. If any of the tests return False
            then the step is rejected. If the latter, then this will override any
            other tests in order to accept the step. This can be used, for example,
            to forcefully escape from a local minimum that ``basinhopping`` is
            trapped in.
        callback : callable, ``callback(x, f, accept)``, optional
            A callback function which will be called for all minima found.  ``x``
            and ``f`` are the coordinates and function value of the trial minimum,
            and ``accept`` is whether or not that minimum was accepted.  This can be
            used, for example, to save the lowest N minima found.  Also,
            ``callback`` can be used to specify a user defined stop criterion by
            optionally returning True to stop the ``basinhopping`` routine.
        interval : integer, optional
            interval for how often to update the ``stepsize``
        disp : bool, optional
            Set to True to print status messages
        niter_success : integer, optional
            Stop the run if the global minimum candidate remains the same for this
            number of iterations.
        seed : int or `np.random.RandomState`, optional
            If `seed` is not specified the `np.RandomState` singleton is used.
            If `seed` is an int, a new `np.random.RandomState` instance is used,
            seeded with seed.
            If `seed` is already a `np.random.RandomState instance`, then that
            `np.random.RandomState` instance is used.
            Specify `seed` for repeatable minimizations. The random numbers
            generated with this seed only affect the default Metropolis
            `accept_test` and the default `take_step`. If you supply your own
            `take_step` and `accept_test`, and these functions use random
            number generation, then those functions are responsible for the state
            of their random number generator.
        init_guess (array, optional): initial guess for the optimization
        init_guess_method (str): method to construct initial guesses if init_guess is not provided.
            Choose from: random
        init_guess_seed (str): random seed for init_guess_method

        Returns
        -------
        res : OptimizeResult
            The optimization result represented as a ``OptimizeResult`` object.  Important
            attributes are: ``x`` the solution array, ``fun`` the value of the
            function at the solution, and ``message`` which describes the cause of
            the termination. The ``OptimzeResult`` object returned by the selected
            minimizer at the lowest minimum is also contained within this object
            and can be accessed through the ``lowest_optimization_result`` attribute.
            See `OptimizeResult` for a description of other attributes.
        See Also
        --------
        minimize :
            The local minimization function called once for each basinhopping step.
            ``minimizer_kwargs`` is passed to this routine.
        Notes
        -----
        Basin-hopping is a stochastic algorithm which attempts to find the global
        minimum of a smooth scalar function of one or more variables [1]_ [2]_ [3]_
        [4]_.  The algorithm in its current form was described by David Wales and
        Jonathan Doye [2]_ http://www-wales.ch.cam.ac.uk/.
        The algorithm is iterative with each cycle composed of the following
        features
        1) random perturbation of the coordinates
        2) local minimization
        3) accept or reject the new coordinates based on the minimized function
           value
        The acceptance test used here is the Metropolis criterion of standard Monte
        Carlo algorithms, although there are many other possibilities [3]_.
        This global minimization method has been shown to be extremely efficient
        for a wide variety of problems in physics and chemistry.  It is
        particularly useful when the function has many minima separated by large
        barriers. See the Cambridge Cluster Database
        http://www-wales.ch.cam.ac.uk/CCD.html for databases of molecular systems
        that have been optimized primarily using basin-hopping.  This database
        includes minimization problems exceeding 300 degrees of freedom.
        See the free software program GMIN (http://www-wales.ch.cam.ac.uk/GMIN) for
        a Fortran implementation of basin-hopping.  This implementation has many
        different variations of the procedure described above, including more
        advanced step taking algorithms and alternate acceptance criterion.
        For stochastic global optimization there is no way to determine if the true
        global minimum has actually been found. Instead, as a consistency check,
        the algorithm can be run from a number of different random starting points
        to ensure the lowest minimum found in each example has converged to the
        global minimum.  For this reason ``basinhopping`` will by default simply
        run for the number of iterations ``niter`` and return the lowest minimum
        found.  It is left to the user to ensure that this is in fact the global
        minimum.
        Choosing ``stepsize``:  This is a crucial parameter in ``basinhopping`` and
        depends on the problem being solved.  Ideally it should be comparable to
        the typical separation between local minima of the function being
        optimized.  ``basinhopping`` will, by default, adjust ``stepsize`` to find
        an optimal value, but this may take many iterations.  You will get quicker
        results if you set a sensible value for ``stepsize``.
        Choosing ``T``: The parameter ``T`` is the temperature used in the
        metropolis criterion.  Basinhopping steps are accepted with probability
        ``1`` if ``func(xnew) < func(xold)``, or otherwise with probability::
            exp( -(func(xnew) - func(xold)) / T )
        So, for best results, ``T`` should to be comparable to the typical
        difference in function values between local minima.
        .. versionadded:: 0.12.0
        References
        ----------
        .. [1] Wales, David J. 2003, Energy Landscapes, Cambridge University Press,
            Cambridge, UK.
        .. [2] Wales, D J, and Doye J P K, Global Optimization by Basin-Hopping and
            the Lowest Energy Structures of Lennard-Jones Clusters Containing up to
            110 Atoms.  Journal of Physical Chemistry A, 1997, 101, 5111.
        .. [3] Li, Z. and Scheraga, H. A., Monte Carlo-minimization approach to the
            multiple-minima problem in protein folding, Proc. Natl. Acad. Sci. USA,
            1987, 84, 6611.
        .. [4] Wales, D. J. and Scheraga, H. A., Global optimization of clusters,
            crystals, and biomolecules, Science, 1999, 285, 1368.
        Examples
        --------
        The following example is a one-dimensional minimization problem,  with many
        local minima superimposed on a parabola.
        >>> from scipy.optimize import basinhopping
        >>> func = lambda x: np.cos(14.5 * x - 0.3) + (x + 0.2) * x
        >>> x0=[1.]
        Basinhopping, internally, uses a local minimization algorithm.  We will use
        the parameter ``minimizer_kwargs`` to tell basinhopping which algorithm to
        use and how to set up that minimizer.  This parameter will be passed to
        ``scipy.optimize.minimize()``.
        >>> minimizer_kwargs = {"method": "BFGS"}
        >>> ret = basinhopping(func, x0, minimizer_kwargs=minimizer_kwargs,
        ...                    niter=200)
        >>> print("global minimum: x = %.4f, f(x0) = %.4f" % (ret.x, ret.fun))
        global minimum: x = -0.1951, f(x0) = -1.0009
        Next consider a two-dimensional minimization problem. Also, this time we
        will use gradient information to significantly speed up the search.
        >>> def func2d(x):
        ...     f = np.cos(14.5 * x[0] - 0.3) + (x[1] + 0.2) * x[1] + (x[0] +
        ...                                                            0.2) * x[0]
        ...     df = np.zeros(2)
        ...     df[0] = -14.5 * np.sin(14.5 * x[0] - 0.3) + 2. * x[0] + 0.2
        ...     df[1] = 2. * x[1] + 0.2
        ...     return f, df
        We'll also use a different local minimization algorithm.  Also we must tell
        the minimizer that our function returns both energy and gradient (jacobian)
        >>> minimizer_kwargs = {"method":"L-BFGS-B", "jac":True}
        >>> x0 = [1.0, 1.0]
        >>> ret = basinhopping(func2d, x0, minimizer_kwargs=minimizer_kwargs,
        ...                    niter=200)
        >>> print("global minimum: x = [%.4f, %.4f], f(x0) = %.4f" % (ret.x[0],
        ...                                                           ret.x[1],
        ...                                                           ret.fun))
        global minimum: x = [-0.1951, -0.1000], f(x0) = -1.0109
        Here is an example using a custom step taking routine.  Imagine you want
        the first coordinate to take larger steps then the rest of the coordinates.
        This can be implemented like so:
        >>> class MyTakeStep(object):
        ...    def __init__(self, stepsize=0.5):
        ...        self.stepsize = stepsize
        ...    def __call__(self, x):
        ...        s = self.stepsize
        ...        x[0] += np.random.uniform(-2.*s, 2.*s)
        ...        x[1:] += np.random.uniform(-s, s, x[1:].shape)
        ...        return x
        Since ``MyTakeStep.stepsize`` exists basinhopping will adjust the magnitude
        of ``stepsize`` to optimize the search.  We'll use the same 2-D function as
        before
        >>> mytakestep = MyTakeStep()
        >>> ret = basinhopping(func2d, x0, minimizer_kwargs=minimizer_kwargs,
        ...                    niter=200, take_step=mytakestep)
        >>> print("global minimum: x = [%.4f, %.4f], f(x0) = %.4f" % (ret.x[0],
        ...                                                           ret.x[1],
        ...                                                           ret.fun))
        global minimum: x = [-0.1951, -0.1000], f(x0) = -1.0109
        Now let's do an example using a custom callback function which prints the
        value of every minimum found
        >>> def print_fun(x, f, accepted):
        ...         print("at minimum %.4f accepted %d" % (f, int(accepted)))
        We'll run it for only 10 basinhopping steps this time.
        >>> np.random.seed(1)
        >>> ret = basinhopping(func2d, x0, minimizer_kwargs=minimizer_kwargs,
        ...                    niter=10, callback=print_fun)
        at minimum 0.4159 accepted 1
        at minimum -0.9073 accepted 1
        at minimum -0.1021 accepted 1
        at minimum -0.1021 accepted 1
        at minimum 0.9102 accepted 1
        at minimum 0.9102 accepted 1
        at minimum 2.2945 accepted 0
        at minimum -0.1021 accepted 1
        at minimum -1.0109 accepted 1
        at minimum -1.0109 accepted 1
        The minimum at -1.0109 is actually the global minimum, found already on the
        8th iteration.
        Now let's implement bounds on the problem using a custom ``accept_test``:
        >>> class MyBounds(object):
        ...     def __init__(self, xmax=[1.1,1.1], xmin=[-1.1,-1.1] ):
        ...         self.xmax = np.array(xmax)
        ...         self.xmin = np.array(xmin)
        ...     def __call__(self, **kwargs):
        ...         x = kwargs["x_new"]
        ...         tmax = bool(np.all(x <= self.xmax))
        ...         tmin = bool(np.all(x >= self.xmin))
        ...         return tmax and tmin
        >>> mybounds = MyBounds()
        >>> ret = basinhopping(func2d, x0, minimizer_kwargs=minimizer_kwargs,
        ...                    niter=10, accept_test=mybounds)
        """

        kwargs = locals().copy()
        self.goal = kwargs['goal']
        for attr in ['self', 'goal', 'init_guess', 'init_guess_method', 'init_guess_seed']:
            del kwargs[attr]

        self.kwargs = kwargs
        self.has_minimizer = False
        self.is_converged  = False
        AbstractPlanner.__init__(**locals())


    def _set_param_space(self, param_space):
        self.param_space = param_space
        self.bounds      = get_bounds(param_space)
        if self.init_guess is None:
            self.init_guess  = get_init_guess(param_space, method=self.init_guess_method, random_seed=self.init_guess_seed)

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        if len(self._values) > 0:
            self.RECEIVED_VALUES.append(self._values[-1])

    def _priv_evaluator(self, params):
        params = self._project_into_domain(params)
        self.SUBMITTED_PARAMS.append(params)
        while len(self.RECEIVED_VALUES) == 0:
            time.sleep(0.1)
        value = self.RECEIVED_VALUES.pop(0)
        return value

    @daemon
    def create_optimizer(self):
        from scipy.optimize import basinhopping
        _ = basinhopping(self._priv_evaluator, x0=self.init_guess, **self.kwargs)
        self.is_converged = True

    def _ask(self):
        if self.has_minimizer is False:
            self.create_optimizer()
            self.has_minimizer = True
        while len(self.SUBMITTED_PARAMS) == 0:
            time.sleep(0.1)
            if self.is_converged:
                return ParameterVector().from_dict(self._params[-1])
        params = self.SUBMITTED_PARAMS.pop(0)
        return ParameterVector().from_array(params, self.param_space)



#!/usr/bin/env python

import numpy as np
from deap import base, creator, tools
from olympus.planners import AbstractPlanner
from olympus.objects import ParameterVector
from olympus import Logger
from copy import deepcopy


# ==========
# Main Class
# ==========
class Genetic(AbstractPlanner):

    def __init__(self, goal='minimize', pop_size=10, cx_prob=0.5, mut_prob=0.2, verbose=False,
                 mate_args={'function':tools.cxTwoPoint},
                 mutate_args={'function':tools.mutGaussian, 'mu':0, 'sigma':0.2, 'indpb':0.2},
                 select_args={'function':tools.selTournament, 'tournsize':3}):
        """Evolutionary Algorithm implemented using the DEAP library.

        We initialize the population using a uniform probability across the parameter space. We then evaluate
        one individual of the population at a time. After all individuals have been evaluated, we generate a set of
        offsprings via selection, cross-over and mutation, using the functions chosen in `select_args`, `mate_args`
        and `mutate_args`. Only novel offsprings, having a set of parameters different from those of the parent
        population, are chosen to be evaluated, and are then proposed sequentially. Once all these offsprings have
        been evaluated, a new set is generated following the same procedure.

        Args:
            goal (str): The optimization goal, either 'minimize' or 'maximize'. Default is 'minimize'.
            pop_size (int): Size of the population. Default is 10.
            cx_prob (float): Probability with which two individuals are crossed. Default is 0.5.
            mut_prob (float): Probability for mutating an individual. Default is 0.2.
            verbose (bool): Whether to print out information. Default is True.
            mate_args (dict): Dictionary containing the details of the crossover function to be used. This should
                contain all arguments that you want to provided to Toolbox.register("mate"). For all crossover
                operations available see https://deap.readthedocs.io/en/master/api/tools.html.
            mutate_args (dict): Dictionary containing the details of the mutation function to be used. This should
                contain all arguments that you want to provided to Toolbox.register("mutate"). For all mutation
                operations available see https://deap.readthedocs.io/en/master/api/tools.html.
            select_args (dict): Dictionary containing the details of the selection function to be used. This should
                contain all arguments that you want to provided to Toolbox.register("select"). For all selection
                operations available see https://deap.readthedocs.io/en/master/api/tools.html.
        """

        AbstractPlanner.__init__(**locals())

        # define fitness function = minimize
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,))  # -1 for minimisation, +1 for maximisation
        creator.create("Individual", list, fitness=creator.FitnessMin)

        # get Deap toolbox
        self.toolbox = base.Toolbox()

    def reset(self):
        """Clears the current population so that a new optimization would start from a randomly initialised population
        again.
        """
        if hasattr(self, 'pop'):
            # reset attributes
            self.num_generated = 0
            self.pop = []
            self.latest_pop_size = None
            self.offsprings = []
            self.novel_offsprings = []
            self.offsprings_to_be_evaluated = []
            # unregister individuals
            self.toolbox.unregister("individual")
            self.toolbox.unregister("population")
            # (re)define fitness function
            creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
            creator.create("Individual", list, fitness=creator.FitnessMin)

    def _register_deap_operations(self):
        """register mate, mutate and select functions"""

        # --------------
        # mate operation
        # --------------
        self.toolbox.register("mate", **self.mate_args)

        # ----------------
        # mutate operation
        # ----------------
        # NOTE: when using default mutate_args, the sigma value is not scaled depending on the range of the parameter.
        # If desired, the scaled sigma values need to be computed outside of this wrapper and passed to the 'sigma'
        # key as a list.
        self.toolbox.register("mutate", **self.mutate_args)

        # ----------------
        # select operation
        # ----------------
        self.toolbox.register("select", **self.select_args)

        # ----------------------------------
        # decorate mate and mutate operations
        # -----------------------------------
        bounds = [param['domain'] for param in self._param_space]
        self.toolbox.decorate("mate", self.project_bounds(bounds))
        self.toolbox.decorate("mutate", self.project_bounds(bounds))

        # Alternative option to control the bounds:
        # self.toolbox.register("mate", tools.cxSimulatedBinaryBounded, low=BOUND_LOW, up=BOUND_UP, eta=20.0)
        # self.toolbox.register("mutate", tools.mutPolynomialBounded, low=BOUND_LOW, up=BOUND_UP, eta=20.0,
        #                       indpb=1.0/len(self._param_space))

    def _set_param_space(self, param_space):
        self._param_space = []
        for param in param_space:
            if param.type == 'continuous':
                param_dict = {'name': param.name, 'type': param.type, 'domain': (param.low, param.high)}
            self._param_space.append(param_dict)

        # register here as they depend on the param_space
        self._register_deap_operations()

    def _tell(self, observations):
        self._params = observations.get_params(as_array=False)
        self._values = observations.get_values(as_array=True, opposite=self.flip_measurements)
        # reshape the array so to have shape (num observations, num objectives)
        # WARNING: In this case we are sticking to 1 objective, multiple objective are not supported!
        # TODO: fix the following reshaping to allow multiple objectives
        self._values = np.reshape(self._values, newshape=(np.shape(self._values)[0], 1))

    def _generate_first_population(self):
        if self.verbose is True:
            Logger.log(f'Creating first population of size {self.pop_size}', 'INFO')

        # Structure initializers
        bounds = [param['domain'] for param in self._param_space]
        self.toolbox.register("individual", self.initIndividual, icls=creator.Individual, bounds=bounds)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)
        self.pop = self.toolbox.population(n=self.pop_size)
        self.latest_pop_size = self.pop_size

        # delete creator classes
        del creator.Individual
        del creator.FitnessMin

    def _set_population_fitness(self):

        # assign evaluations to each offspring
        offsprings_fitness = self._values[-self.latest_pop_size:, :]
        assert len(self.novel_offsprings) == len(offsprings_fitness)

        # Note we operate on self.novel_offsprings, but this will also update self.offsprings, which is the population
        # used for selection, cross-over and mutation
        for ind, fit in zip(self.novel_offsprings, offsprings_fitness):
            if not ind.fitness.valid:
                ind.fitness.values = (fit[0],)

        # replace the old population by the offsprings.
        self.pop[:] = self.offsprings

    def _generate_offsprings(self):
        if self.verbose is True:
            Logger.log('Creating new offsprings...', 'INFO')

        # Select the next generation individuals
        self.offsprings = self.toolbox.select(self.pop, len(self.pop))
        # Clone the selected individuals
        self.offsprings = list(map(self.toolbox.clone, self.offsprings))

        # Apply crossover and on the offsprings
        for child1, child2 in zip(self.offsprings[::2], self.offsprings[1::2]):
            if np.random.uniform() < self.cx_prob:
                if self.verbose is True:
                    Logger.log('  Performing cross-over operation', 'INFO')
                self.toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values

        # Apply mutation on the offsprings
        for mutant in self.offsprings:
            if np.random.uniform() < self.mut_prob:
                if self.verbose is True:
                    Logger.log('  Performing mutation operation', 'INFO')
                self.toolbox.mutate(mutant)
                del mutant.fitness.values

        # Evaluate the individuals with an invalid fitness, i.e. those that have been mutated. The rest have not
        # changed ==> no need to re-evaluate them
        self.novel_offsprings = [ind for ind in self.offsprings if not ind.fitness.valid]
        if self.verbose is True:
            Logger.log(f'  {len(self.novel_offsprings)} novel offsprings found', 'INFO')

        # use only novel offsprings for evaluation
        # for the others we already know the outcome - no need to re-evaluate them
        self.offsprings_to_be_evaluated = deepcopy(self.novel_offsprings)
        self.latest_pop_size = len(self.offsprings_to_be_evaluated)

    def _ask(self):

        if self.num_generated == 0:
            raise NotImplementedError('the attribute "num_generated" was not expected to ever be zero here!')

        # If it is the first generation (note num_generated index starts from 1) we initialize the population
        if self.num_generated < 2:
            self._generate_first_population()
            self.offsprings = self.pop  # first iteration, so there is no offspring yet
            self.novel_offsprings = self.offsprings  # all offsprings are novel
            self.offsprings_to_be_evaluated = deepcopy(self.novel_offsprings)
        # otherwise select, mate and mutate population
        else:
            if len(self.offsprings_to_be_evaluated) == 0:
                # update population with previous offsprings and their associated observations/evaluations
                self._set_population_fitness()
                # cross-over and mutate, then select new offsprings to be evaluated
                # keep doing this if by chance we did not do any cross-over or mutation
                while len(self.offsprings_to_be_evaluated) == 0:
                    self._generate_offsprings()

        # evaluate one offspring at a time
        array = self.offsprings_to_be_evaluated.pop(0)
        return ParameterVector().from_array(array, self.param_space)

    @staticmethod
    def initIndividual(icls, bounds):
        """attribute generator: this defines how we initialize the population"""
        return icls(np.random.uniform(*p) for p in bounds)

    @staticmethod
    def project_bounds(bounds):
        """DEAP decorator to project out of bounds parameters back onto the boundary.
        `bounds` looks like this: [(low, high), (low, high), (low, high)]
        """
        def decorator(func):
            def wrapper(*args, **kargs):
                offspring = func(*args, **kargs)
                for child in offspring:
                    assert len(child) == len(bounds)
                    for i in range(len(child)):
                        if child[i] > bounds[i][1]:
                            child[i] = bounds[i][1]
                        elif child[i] < bounds[i][0]:
                            child[i] = bounds[i][0]
                return offspring
            return wrapper
        return decorator




#!/usr/bin/env python

#===============================================================================

import pickle

#===============================================================================

from olympus.databases import AbstractDatabase

#===============================================================================

class Wrapper_pickle(AbstractDatabase):

    ''' implements a simple database using pickle

    Architecture of the database:
        giant dictionary where keys are the identifiers of individual campaigns
        and the associated attributes represent the campaign as yet another
        dictionary
    '''

    ATT_KIND      = {'type': 'string', 'default': 'pkl'}
    ATT_FILE_TYPE = {'type': 'string', 'default': 'pkl', 'valid': ['pkl', 'pickle']}


    def _load_db(self):
        # ATTENTION: requires a lock if we ever run into race conditions
        with open(self.file_name, 'rb') as content:
            db = pickle.load(content)
        self.db = db
        return self


    def _list_campaign_ids(self):
        self._load_db()
        return list(self.db.keys())


    def _log_campaign(self, campaign):
        if self.db_exists:
            self._load_db()
        else:
            self.db = {}
        self.db[campaign.id] = campaign.to_dict()
        self._save_db(self.db)


    def _save_db(self, db):
        with open(self.file_name, 'wb') as content:
            pickle.dump(db, content)





#===============================================================================
